import io
import json
import pathlib

from typing import Callable, List, Optional, TypeVar, Union, cast
from typing_extensions import ParamSpec

from weaviate.collections.classes.aggregate import (
    PropertiesMetrics,
    AProperties,
    AggregateResult,
    AggregateBool,
    AggregateDate,
    AggregateInt,
    AggregateFloat,
    AggregateRef,
    AggregateStr,
    _AggregateGroupByReturn,
    _AggregateReturn,
    _Metrics,
    _MetricsBase,
    _MetricsBool,
    _MetricsDate,
    _MetricsFloat,
    _MetricsInt,
    _MetricsRef,
    _MetricsStr,
    _GroupedBy,
    TopOccurrence,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.classes.grpc import Move
from weaviate.connect import Connection
from weaviate.collections.filters import _FilterToREST
from weaviate.exceptions import WeaviateInvalidInputException, WeaviateQueryException
from weaviate.gql.aggregate import AggregateBuilder
from weaviate.util import file_encoder_b64
from weaviate.types import UUID

P = ParamSpec("P")
T = TypeVar("T")


def validate(fn: Callable[P, T]) -> Callable[P, T]:
    """Validate the aggregations argument."""

    def inner(*args: P.args, **kwargs: P.kwargs) -> T:
        try:
            aggregations = args[1]
        except IndexError:
            aggregations = kwargs.get("return_metrics")
        if not isinstance(aggregations, list):
            raise TypeError(
                f"The aggregations argument received an unexpected type: {type(aggregations)}. This argument must be a list!"
            )
        for aggregation in aggregations:
            if not isinstance(aggregation, _MetricsBase):
                raise TypeError(
                    f"One of the aggregations is an unexpected type: {type(aggregation)}. Did you forget to append .returning() to .with_()?"
                )
        return fn(*args, **kwargs)

    return inner


class _Aggregate:
    def __init__(
        self,
        connection: Connection,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
    ):
        self.__connection = connection
        self.__name = name
        self._tenant = tenant
        self._consistency_level = consistency_level

    def _query(self) -> AggregateBuilder:
        return AggregateBuilder(self.__name, self.__connection)

    def _to_aggregate_result(
        self, response: dict, metrics: Optional[PropertiesMetrics]
    ) -> _AggregateReturn:
        try:
            result: dict = response["data"]["Aggregate"][self.__name][0]
            return _AggregateReturn(
                properties=self.__parse_properties(result, metrics) if metrics is not None else {},
                total_count=result["meta"]["count"] if result.get("meta") is not None else None,
            )
        except KeyError as e:
            raise ValueError(
                f"There was an error accessing the {e} key when parsing the GraphQL response: {response}"
            )

    def _to_group_by_result(
        self, response: dict, metrics: Optional[PropertiesMetrics]
    ) -> List[_AggregateGroupByReturn]:
        try:
            results: dict = response["data"]["Aggregate"][self.__name]
            return [
                _AggregateGroupByReturn(
                    grouped_by=_GroupedBy(
                        prop=result["groupedBy"]["path"][0],
                        value=result["groupedBy"]["value"],
                    ),
                    properties=self.__parse_properties(result, metrics)
                    if metrics is not None
                    else {},
                    total_count=result["meta"]["count"] if result.get("meta") is not None else None,
                )
                for result in results
            ]
        except KeyError as e:
            raise ValueError(
                f"There was an error accessing the {e} key when parsing the GraphQL response: {response}"
            )

    def __parse_properties(self, result: dict, metrics: PropertiesMetrics) -> AProperties:
        props: AProperties = {}
        for metric in metrics:
            if metric.property_name in result:
                props[metric.property_name] = self.__parse_property(
                    result[metric.property_name], metric
                )
        return props

    @staticmethod
    def __parse_property(property_: dict, metric: _Metrics) -> AggregateResult:
        if isinstance(metric, _MetricsStr):
            return AggregateStr(
                count=property_.get("count"),
                top_occurrences=[
                    TopOccurrence(
                        occurs=cast(dict, top_occurence).get("occurs"),
                        value=cast(dict, top_occurence).get("value"),
                    )
                    for top_occurence in property_.get("topOccurrences", [])
                ],
            )
        elif isinstance(metric, _MetricsInt):
            return AggregateInt(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                mean=property_.get("mean"),
                median=property_.get("median"),
                mode=property_.get("mode"),
                sum_=property_.get("sum"),
            )
        elif isinstance(metric, _MetricsFloat):
            return AggregateFloat(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                mean=property_.get("mean"),
                median=property_.get("median"),
                mode=property_.get("mode"),
                sum_=property_.get("sum"),
            )
        elif isinstance(metric, _MetricsBool):
            return AggregateBool(
                count=property_.get("count"),
                percentage_false=property_.get("percentageFalse"),
                percentage_true=property_.get("percentageTrue"),
                total_false=property_.get("totalFalse"),
                total_true=property_.get("totalTrue"),
            )
        elif isinstance(metric, _MetricsDate):
            return AggregateDate(
                count=property_.get("count"),
                maximum=property_.get("maximum"),
                median=property_.get("median"),
                minimum=property_.get("minimum"),
                mode=property_.get("mode"),
            )
        elif isinstance(metric, _MetricsRef):
            return AggregateRef(pointing_to=property_.get("pointingTo"))
        else:
            raise ValueError(
                f"Unknown aggregation type {metric} encountered in _Aggregate.__parse_property() for property {property_}"
            )

    def _base(
        self,
        metrics: Optional[PropertiesMetrics],
        filters: Optional[_Filters],
        limit: Optional[int],
        total_count: bool,
    ) -> AggregateBuilder:
        builder = self._query()
        if metrics is not None:
            builder = builder.with_fields(" ".join([metric.to_gql() for metric in metrics]))
        if filters is not None:
            builder = builder.with_where(_FilterToREST.convert(filters))
        if limit is not None:
            builder = builder.with_limit(limit)
        if total_count:
            builder = builder.with_meta_count()
        if self._tenant is not None:
            builder = builder.with_tenant(self._tenant)
        return builder

    @staticmethod
    def _do(query: AggregateBuilder) -> dict:
        res = query.do()
        if res.get("errors") is not None:
            raise WeaviateQueryException(
                f"Error in GraphQL response: {json.dumps(res['errors'], indent=2)}, for the following query: {query.build()}"
            )
        return res

    @staticmethod
    def _add_near_image(
        builder: AggregateBuilder,
        near_image: Union[str, pathlib.Path, io.BufferedReader],
        certainty: Optional[float],
        distance: Optional[float],
        object_limit: Optional[int],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputException(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        payload: dict = {}
        payload["image"] = _parse_media(near_image)
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        builder = builder.with_near_image(payload, encode=False)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_object(
        builder: AggregateBuilder,
        near_object: UUID,
        certainty: Optional[float],
        distance: Optional[float],
        object_limit: Optional[int],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputException(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        payload: dict = {}
        payload["id"] = str(near_object)
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        builder = builder.with_near_object(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_text(
        builder: AggregateBuilder,
        query: Union[List[str], str],
        certainty: Optional[float],
        distance: Optional[float],
        move_to: Optional[Move],
        move_away: Optional[Move],
        object_limit: Optional[int],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputException(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        payload: dict = {}
        payload["concepts"] = query if isinstance(query, list) else [query]
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        if move_to is not None:
            payload["moveTo"] = move_to._to_gql_payload()
        if move_away is not None:
            payload["moveAwayFrom"] = move_away._to_gql_payload()
        builder = builder.with_near_text(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder

    @staticmethod
    def _add_near_vector(
        builder: AggregateBuilder,
        near_vector: List[float],
        certainty: Optional[float],
        distance: Optional[float],
        object_limit: Optional[int],
    ) -> AggregateBuilder:
        if all([certainty is None, distance is None, object_limit is None]):
            raise WeaviateInvalidInputException(
                "You must provide at least one of the following arguments: certainty, distance, object_limit when vector searching"
            )
        payload: dict = {}
        payload["vector"] = near_vector
        if certainty is not None:
            payload["certainty"] = certainty
        if distance is not None:
            payload["distance"] = distance
        builder = builder.with_near_vector(payload)
        if object_limit is not None:
            builder = builder.with_object_limit(object_limit)
        return builder


def _parse_media(media: Union[str, pathlib.Path, io.BufferedReader]) -> str:
    if isinstance(media, str):  # if already encoded by user
        return media
    else:
        return file_encoder_b64(media)



"""
Weaviate Python Client Library used to interact with a Weaviate instance.
"""


__all__ = [
    "Client",
    "Connect",
    "WeaviateClient",
    "AuthClientCredentials",
    "AuthClientPassword",
    "AuthBearerToken",
    "AuthApiKey",
    "UnexpectedStatusCodeException",
    "ObjectAlreadyExistsException",
    "AuthenticationFailedException",
    "SchemaValidationException",
    "WeaviateStartUpError",
    "ConsistencyLevel",
    "WeaviateErrorRetryConf",
    "EmbeddedOptions",
    "AdditionalConfig",
    "Config",
    "ConnectionConfig",
    "ConnectionParams",
    "ProtocolParams",
    "AdditionalProperties",
    "LinkTo",
    "Tenant",
    "TenantActivityStatus",
    "connect_to_custom",
    "connect_to_embedded",
    "connect_to_local",
    "connect_to_wcs",
]

import sys

from importlib.metadata import version, PackageNotFoundError

try:
    __version__ = version("weaviate-client")
except PackageNotFoundError:
    __version__ = "unknown version"

from .auth import AuthClientCredentials, AuthClientPassword, AuthBearerToken, AuthApiKey
from .batch.crud_batch import WeaviateErrorRetryConf
from .client import Client, WeaviateClient
from .connect.connection import ConnectionParams, ProtocolParams
from .data.replication import ConsistencyLevel
from .schema.crud_schema import Tenant, TenantActivityStatus
from .connect.helpers import (
    connect_to_custom,
    connect_to_embedded,
    connect_to_local,
    connect_to_wcs,
)
from .embedded import EmbeddedOptions
from .exceptions import (
    UnexpectedStatusCodeException,
    ObjectAlreadyExistsException,
    AuthenticationFailedException,
    SchemaValidationException,
    WeaviateStartUpError,
)
from .config import AdditionalConfig, Config, ConnectionConfig
from .gql.get import AdditionalProperties, LinkTo

if not sys.warnoptions:
    import warnings

    warnings.simplefilter("default")



import datetime
from dataclasses import dataclass, field
from typing import List, Dict, Optional, Any

import pytest
import uuid

import weaviate

schema = {
    "classes": [
        {
            "class": "Paragraph",
            "properties": [
                {"dataType": ["text"], "name": "contents"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
                {"dataType": ["Author"], "name": "author"},
            ],
            "vectorizer": "none",
        },
        {
            "class": "Article",
            "properties": [
                {"dataType": ["string"], "name": "title"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
                {"dataType": ["date"], "name": "datePublished"},
                {"dataType": ["Author"], "name": "author"},
                {"dataType": ["string"], "name": "somestring"},
                {"dataType": ["int"], "name": "counter"},
            ],
            "vectorizer": "none",
        },
        {
            "class": "Author",
            "properties": [{"dataType": ["string"], "name": "name"}],
            "vectorizer": "none",
        },
    ]
}


@dataclass(frozen=True)
class Reference:
    to_class: str
    to_uuid: uuid.UUID


@dataclass
class DataObject:
    properties: Dict[str, Any]
    class_name: str
    uuid: uuid


@dataclass
class Author:
    name: str
    uuid: uuid = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject({"name": self.name}, self.class_name, self.uuid)

    def __post_init__(self) -> None:
        self.uuid = uuid.uuid4()
        self.class_name = "Author"


@dataclass
class Paragraph:
    contents: str
    author: Reference
    hasParagraphs: Optional[Reference]
    uuid: uuid = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject({"contents": self.contents}, self.class_name, self.uuid)

    def __post_init__(self) -> None:
        self.uuid = uuid.uuid4()
        self.class_name = "Paragraph"


@dataclass
class Article:
    title: str
    datePublished: str
    somestring: str
    counter: int
    author: Reference
    hasParagraphs: Reference
    uuid: uuid = field(init=False)
    class_name: str = field(init=False)

    def to_data_object(self) -> DataObject:
        return DataObject(
            {"title": self.title, "datePublished": self.datePublished}, self.class_name, self.uuid
        )

    def __post_init__(self) -> None:
        self.uuid = uuid.uuid4()
        self.class_name = "Article"


@pytest.mark.parametrize("dynamic", [False])
@pytest.mark.parametrize("batch_size", [50])
def test_stress(batch_size, dynamic):
    client = weaviate.Client("http://localhost:8080")
    client.schema.delete_all()
    client.schema.create(schema)
    client.batch.configure(batch_size=batch_size, dynamic=dynamic, num_workers=4)
    authors = create_authors(1000)
    paragraphs = create_paragraphs(1000, authors)
    articles = create_articles(1000, authors, paragraphs)

    add_authors(client, authors)
    add_paragraphs(client, paragraphs)
    add_articles(client, articles)

    client.batch.flush()
    __assert_add(client, authors, authors[0].class_name)
    __assert_add(client, paragraphs, paragraphs[0].class_name)
    __assert_add(client, articles, articles[0].class_name)

    # verify references
    for article in articles:
        article_weav = client.data_object.get_by_id(article.uuid, class_name=article.class_name)
        beacon_article = str(article_weav["properties"]["author"][0]["beacon"])
        assert beacon_article.split("/")[-1] == str(article.author.to_uuid)
        beacon_paragraph = str(article_weav["properties"]["hasParagraphs"][0]["beacon"])
        assert beacon_paragraph.split("/")[-1] == str(article.hasParagraphs.to_uuid)

    for paragraph in paragraphs:
        article_weav = client.data_object.get_by_id(paragraph.uuid, class_name=paragraph.class_name)
        beacon_article = str(article_weav["properties"]["author"][0]["beacon"])
        assert beacon_article.split("/")[-1] == str(paragraph.author.to_uuid)
        if paragraph.hasParagraphs is not None:
            beacon_paragraph = str(article_weav["properties"]["hasParagraphs"][0]["beacon"])
            assert beacon_paragraph.split("/")[-1] == str(paragraph.hasParagraphs.to_uuid)
        else:
            assert "hasParagraphs" not in article_weav["properties"]

    client.schema.delete_all()


@pytest.fixture(
    params=[(batch_size, workers) for workers in [1, 4, 10] for batch_size in [-1, 50, 100]],
    ids=[
        f"batch_size{batch_size}, workers {workers})"
        for workers in [1, 4, 10]
        for batch_size in [-1, 50, 100]
    ],
)
def client(request):
    local_client = weaviate.Client("http://localhost:8080")
    if request.param[0] > 0:
        local_client.batch.configure(
            batch_size=request.param[0], dynamic=False, num_workers=request.param[1]
        )
    else:
        local_client.batch.configure(batch_size=10, dynamic=True, num_workers=request.param[1])
    return local_client


def run_stress_test(client):
    client.schema.delete_all()
    client.schema.create(schema)

    authors = create_authors(20000)
    paragraphs = create_paragraphs(20000, authors)
    articles = create_articles(10000, authors, paragraphs)

    add_authors(client, authors)
    add_paragraphs(client, paragraphs)
    add_articles(client, articles)

    client.batch.flush()
    __assert_add(client, authors, authors[0].class_name)
    __assert_add(client, paragraphs, paragraphs[0].class_name)
    __assert_add(client, articles, articles[0].class_name)

    client.schema.delete_all()


@pytest.mark.profiling
def test_profile_stress(client):
    run_stress_test(client)


def test_benchmark_stress_test(benchmark, client):
    benchmark(test_profile_stress, client)


def add_authors(client: weaviate.Client, authors: List[Author]):
    for author in authors:
        data_object = author.to_data_object()
        client.batch.add_data_object(
            data_object.properties, data_object.class_name, data_object.uuid
        )


def add_paragraphs(client: weaviate.Client, paragraphs: List[Paragraph]):
    for paragraph in paragraphs:
        data_object = paragraph.to_data_object()
        client.batch.add_data_object(
            data_object.properties, data_object.class_name, data_object.uuid
        )
        client.batch.add_reference(
            str(paragraph.uuid),
            from_property_name="author",
            to_object_uuid=str(paragraph.author.to_uuid),
            from_object_class_name="Paragraph",
            to_object_class_name="Author",
        )
        if paragraph.hasParagraphs is not None:
            client.batch.add_reference(
                str(paragraph.uuid),
                from_property_name="hasParagraphs",
                to_object_uuid=str(paragraph.hasParagraphs.to_uuid),
                from_object_class_name="Paragraph",
                to_object_class_name="Paragraph",
            )


def add_articles(client: weaviate.Client, articles: List[Article]):
    for article in articles:
        data_object = article.to_data_object()
        client.batch.add_data_object(
            data_object.properties, data_object.class_name, data_object.uuid
        )
        client.batch.add_reference(
            str(article.uuid),
            from_property_name="author",
            to_object_uuid=str(article.author.to_uuid),
            from_object_class_name="Article",
            to_object_class_name="Author",
        )
        client.batch.add_reference(
            str(article.uuid),
            from_property_name="hasParagraphs",
            to_object_uuid=str(article.hasParagraphs.to_uuid),
            from_object_class_name="Article",
            to_object_class_name="Paragraph",
        )


def create_authors(num_authors: int) -> List[Author]:
    authors: List[Author] = [Author(f"{i}") for i in range(num_authors)]
    return authors


def create_paragraphs(num_paragraphs: int, authors: List[Author]) -> List[Paragraph]:
    paragraphs: List[Paragraph] = []
    for i in range(num_paragraphs):
        content: str = f"{i} {i} {i} {i}"

        paragraph_to_reference: Optional[Paragraph] = None
        if len(paragraphs) > 0 and i % 2 == 0:
            paragraph_to_reference: Paragraph = paragraphs[i % len(paragraphs)]
        author_to_reference: Author = authors[i % len(authors)]
        paragraphs.append(
            Paragraph(
                content,
                Reference("Author", author_to_reference.uuid),
                Reference("Paragraph", paragraph_to_reference.uuid)
                if paragraph_to_reference is not None
                else None,
            )
        )
    return paragraphs


def create_articles(
    num_articles: int, authors: List[Author], paragraphs: List[Paragraph]
) -> List[Article]:
    articles: List[Article] = []
    base_date: datetime.date = datetime.datetime(2023, 12, 9, 7, 1, 34)
    for i in range(num_articles):
        title: str = f"{i} {i} {i}"
        paragraph_to_reference: Paragraph = paragraphs[i % len(paragraphs)]
        author_to_reference: Author = authors[i % len(authors)]
        date_published: str = (base_date + datetime.timedelta(hours=i)).isoformat() + "Z"
        articles.append(
            Article(
                title,
                date_published,
                str(i),
                i,
                Reference("Author", author_to_reference.uuid),
                Reference("Paragraph", paragraph_to_reference.uuid),
            )
        )

    return articles


def __assert_add(client: weaviate.Client, objects: List[Any], class_name: str) -> None:
    result = client.query.aggregate(class_name).with_meta_count().do()
    assert len(objects) == result["data"]["Aggregate"][class_name][0]["meta"]["count"]



from typing import Optional, Tuple, List

from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.connect import Connection
from weaviate.proto.v1 import base_pb2


class _BaseGRPC:
    def __init__(self, connection: Connection, consistency_level: Optional[ConsistencyLevel]):
        self._connection = connection
        self._consistency_level = self._get_consistency_level(consistency_level)

    def _get_metadata(self) -> Optional[Tuple[Tuple[str, str], ...]]:
        metadata: Optional[Tuple[Tuple[str, str], ...]] = None
        access_token = self._connection.get_current_bearer_token()

        metadata_list: List[Tuple[str, str]] = []
        if len(access_token) > 0:
            metadata_list.append(("authorization", access_token))

        if len(self._connection.additional_headers):
            for key, val in self._connection.additional_headers.items():
                if val is not None:
                    metadata_list.append((key.lower(), val))

        if len(metadata_list) > 0:
            metadata = tuple(metadata_list)

        return metadata

    @staticmethod
    def _get_consistency_level(
        consistency_level: Optional[ConsistencyLevel],
    ) -> Optional["base_pb2.ConsistencyLevel"]:
        if consistency_level is None:
            return None

        if consistency_level.value == ConsistencyLevel.ONE:
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_ONE
        elif consistency_level.value == ConsistencyLevel.QUORUM:
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_QUORUM
        else:
            assert consistency_level.value == ConsistencyLevel.ALL
            return base_pb2.ConsistencyLevel.CONSISTENCY_LEVEL_ALL



from typing import Optional

import pytest
import requests

import weaviate
from weaviate import Tenant, TenantActivityStatus


@pytest.fixture(scope="module")
def client():
    client = weaviate.Client("http://localhost:8080")
    yield client
    client.schema.delete_all()


@pytest.mark.parametrize("replicationFactor", [None, 1])
def test_create_class_with_implicit_and_explicit_replication_factor(
    client: weaviate.Client, replicationFactor: Optional[int]
):
    single_class = {
        "class": "Barbecue",
        "description": "Barbecue or BBQ where meat and vegetables get grilled",
        "properties": [
            {
                "dataType": ["string"],
                "description": "how hot is the BBQ in C",
                "name": "heat",
            },
        ],
    }
    if replicationFactor is None:
        expected_factor = 1
    else:
        expected_factor = replicationFactor
        single_class["replicationConfig"] = {
            "factor": replicationFactor,
        }

    client.schema.create_class(single_class)
    created_class = client.schema.get("Barbecue")
    assert created_class["class"] == "Barbecue"
    assert created_class["replicationConfig"]["factor"] == expected_factor

    client.schema.delete_class("Barbecue")


@pytest.mark.parametrize("data_type", ["uuid", "uuid[]"])
def test_uuid_datatype(client: weaviate.Client, data_type: str):
    single_class = {"class": "UuidTest", "properties": [{"dataType": [data_type], "name": "heat"}]}

    client.schema.create_class(single_class)
    created_class = client.schema.get("uuidTest")
    assert created_class["class"] == "UuidTest"

    client.schema.delete_class("UuidTest")


@pytest.mark.parametrize("object_", ["object", "object[]"])
@pytest.mark.parametrize(
    "nested",
    [
        {
            "dataType": ["text"],
            "name": "name",
        },
        {"dataType": ["text[]"], "name": "names"},
        {"dataType": ["int"], "name": "age"},
        {"dataType": ["int[]"], "name": "ages"},
        {"dataType": ["number"], "name": "weight"},
        {"dataType": ["number[]"], "name": "weights"},
        {"dataType": ["boolean"], "name": "isAlive"},
        {"dataType": ["boolean[]"], "name": "areAlive"},
        {"dataType": ["date"], "name": "birthDate"},
        {"dataType": ["date[]"], "name": "birthDates"},
        {"dataType": ["uuid"], "name": "uuid"},
        {"dataType": ["uuid[]"], "name": "uuids"},
        {"dataType": ["blob"], "name": "blob"},
        {
            "dataType": ["object"],
            "name": "object",
            "nestedProperties": [{"dataType": ["text"], "name": "name"}],
        },
        {
            "dataType": ["object[]"],
            "name": "objects",
            "nestedProperties": [{"dataType": ["text"], "name": "name"}],
        },
    ],
)
def test_object_datatype(client: weaviate.Client, object_: str, nested: dict):
    single_class = {
        "class": "ObjectTest",
        "properties": [{"dataType": [object_], "name": "heat", "nestedProperties": [nested]}],
    }

    client.schema.create_class(single_class)
    created_class = client.schema.get("ObjectTest")
    assert created_class["class"] == "ObjectTest"

    client.schema.delete_class("ObjectTest")


@pytest.mark.parametrize("tokenization", ["word", "whitespace", "lowercase", "field"])
def test_tokenization(client: weaviate.Client, tokenization):
    single_class = {
        "class": "TokenTest",
        "properties": [{"dataType": ["text"], "name": "heat", "tokenization": tokenization}],
    }
    client.schema.create_class(single_class)
    created_class = client.schema.get("TokenTest")
    assert created_class["class"] == "TokenTest"

    client.schema.delete_class("TokenTest")


def test_class_exists(client: weaviate.Client):
    single_class = {"class": "Exists"}

    client.schema.create_class(single_class)
    assert client.schema.exists("Exists") is True
    assert client.schema.exists("DoesNotExists") is False

    client.schema.delete_class("Exists")
    assert client.schema.exists("Exists") is False


def test_schema_keys(client: weaviate.Client):
    single_class = {
        "class": "Author",
        "properties": [
            {
                "indexFilterable": False,
                "indexSearchable": False,
                "dataType": ["text"],
                "name": "name",
            }
        ],
    }
    client.schema.create_class(single_class)
    assert client.schema.exists("Author")


def test_class_tenants(client: weaviate.Client):
    class_name = "MultiTenancySchemaTest"
    uncap_class_name = "multiTenancySchemaTest"
    single_class = {"class": class_name, "multiTenancyConfig": {"enabled": True}}
    client.schema.delete_all()
    client.schema.create_class(single_class)
    assert client.schema.exists(class_name)

    tenants = [
        Tenant(name="Tenant1"),
        Tenant(name="Tenant2"),
        Tenant(name="Tenant3"),
        Tenant(name="Tenant4"),
    ]
    client.schema.add_class_tenants(class_name, tenants[:2])
    client.schema.add_class_tenants(uncap_class_name, tenants[2:])
    tenants_get = client.schema.get_class_tenants(class_name)
    assert len(tenants_get) == len(tenants)

    client.schema.remove_class_tenants(class_name, ["Tenant2", "Tenant4"])
    client.schema.remove_class_tenants(uncap_class_name, ["Tenant1"])
    tenants_get = client.schema.get_class_tenants(uncap_class_name)
    assert len(tenants_get) == 1


def test_update_schema_with_no_properties(client: weaviate.Client):
    single_class = {"class": "NoProperties"}

    requests.post("http://localhost:8080/v1/schema", json=single_class)
    assert client.schema.exists("NoProperties")

    client.schema.update_config("NoProperties", {"vectorIndexConfig": {"ef": 64}})
    assert client.schema.exists("NoProperties")

    client.schema.delete_class("NoProperties")
    assert client.schema.exists("NoProperties") is False


def test_class_tenants_activate_deactivate(client: weaviate.Client):
    class_name = "MultiTenancyActivateDeactivateSchemaTest"
    uncap_class_name = "multiTenancyActivateDeactivateSchemaTest"
    single_class = {"class": class_name, "multiTenancyConfig": {"enabled": True}}
    client.schema.delete_all()
    client.schema.create_class(single_class)
    assert client.schema.exists(class_name)

    tenants = [
        Tenant(name="Tenant1"),
        Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant2"),
        Tenant(name="Tenant3"),
    ]
    client.schema.add_class_tenants(class_name, tenants)
    tenants_get = client.schema.get_class_tenants(class_name)
    assert len(tenants_get) == len(tenants)
    # below required because tenants are returned in random order by the server
    for tenant in tenants_get:
        if tenant.name == "Tenant1":
            assert tenant.activity_status == TenantActivityStatus.HOT
        elif tenant.name == "Tenant2":
            assert tenant.activity_status == TenantActivityStatus.COLD
        elif tenant.name == "Tenant3":
            assert tenant.activity_status == TenantActivityStatus.HOT
        else:
            raise AssertionError(f"Unexpected tenant: {tenant.name}")

    updated_tenants = [
        Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant1"),
        Tenant(activity_status=TenantActivityStatus.HOT, name="Tenant2"),
    ]
    client.schema.update_class_tenants(class_name, updated_tenants)
    tenants_get = client.schema.get_class_tenants(class_name)
    assert len(tenants_get) == len(tenants)
    # below required because tenants are returned in random order by the server
    for tenant in tenants_get:
        if tenant.name == "Tenant1":
            assert tenant.activity_status == TenantActivityStatus.COLD
        elif tenant.name == "Tenant2":
            assert tenant.activity_status == TenantActivityStatus.HOT
        elif tenant.name == "Tenant3":
            assert tenant.activity_status == TenantActivityStatus.HOT
        else:
            raise AssertionError(f"Unexpected tenant: {tenant.name}")

    updated_tenants = [
        Tenant(activity_status=TenantActivityStatus.COLD, name="Tenant3"),
    ]
    client.schema.update_class_tenants(uncap_class_name, updated_tenants)
    tenants_get = client.schema.get_class_tenants(uncap_class_name)
    assert len(tenants_get) == len(tenants)
    # below required because tenants are returned in random order by the server
    for tenant in tenants_get:
        if tenant.name == "Tenant1":
            assert tenant.activity_status == TenantActivityStatus.COLD
        elif tenant.name == "Tenant2":
            assert tenant.activity_status == TenantActivityStatus.HOT
        elif tenant.name == "Tenant3":
            assert tenant.activity_status == TenantActivityStatus.COLD
        else:
            raise AssertionError(f"Unexpected tenant: {tenant.name}")



import unittest
from unittest.mock import Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
)
from weaviate.schema.properties import Property


class TestCRUDProperty(unittest.TestCase):
    def test_create(self):
        """
        Test `create` method.
        """

        prop = Property(Mock())

        # invalid calls
        error_message = "Class name must be of type str but is "
        requests_error_message = "Property was created properly."

        with self.assertRaises(TypeError) as error:
            prop.create(35, {})
        check_error_message(self, error, error_message + str(int))

        prop = Property(mock_connection_func("post", side_effect=RequestsConnectionError("Test!")))
        with self.assertRaises(RequestsConnectionError) as error:
            prop.create("Class", {"name": "test", "dataType": ["test_type"]})
        check_error_message(self, error, requests_error_message)

        prop = Property(mock_connection_func("post", status_code=404))
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            prop.create("Class", {"name": "test", "dataType": ["test_type"]})
        check_startswith_error_message(self, error, "Add property to class")

        # valid calls
        connection_mock = mock_connection_func("post")  # Mock calling weaviate
        prop = Property(connection_mock)

        test_prop = {
            "dataType": ["string"],
            "description": "my Property",
            "moduleConfig": {"text2vec-contextionary": {"vectorizePropertyName": True}},
            "name": "superProp",
            "indexInverted": True,
        }

        prop.create("TestThing", test_prop)

        connection_mock.post.assert_called_with(
            path="/schema/TestThing/properties",
            weaviate_object=test_prop,
        )

        prop.create("testThing", test_prop)

        connection_mock.post.assert_called_with(
            path="/schema/TestThing/properties",
            weaviate_object=test_prop,
        )



import uuid as uuid_package
from dataclasses import dataclass
from typing import Any, Dict, List, Optional, Sequence, Union, cast

from pydantic import BaseModel, Field, field_validator

from weaviate.util import _capitalize_first_letter, get_valid_uuid, get_vector
from weaviate.types import BEACON, UUID, WeaviateField


@dataclass
class _BatchObject:
    class_name: str
    vector: Optional[List[float]]
    uuid: Optional[UUID]
    properties: Dict[str, WeaviateField]
    tenant: Optional[str]


@dataclass
class _BatchReference:
    from_: str
    to: str
    tenant: Optional[str]


class BatchObject(BaseModel):
    """
    A Weaviate object to be added to the database.

    Performs validation on the class name and UUID, and automatically generates a UUID if one is not provided.
    Also converts the vector to a list of floats if it is provided as a numpy array.
    """

    class_name: str
    properties: Dict[str, Any]
    uuid: Optional[UUID] = Field(default=None)
    vector: Optional[Sequence] = Field(default=None)
    tenant: Optional[str] = Field(default=None)

    def __init__(self, **data: Any) -> None:
        data["vector"] = get_vector(v) if (v := data.get("vector")) is not None else None
        data["uuid"] = (
            get_valid_uuid(u) if (u := data.get("uuid")) is not None else uuid_package.uuid4()
        )
        super().__init__(**data)

    def _to_internal(self) -> _BatchObject:
        return _BatchObject(
            class_name=self.class_name,
            vector=cast(list, self.vector),
            uuid=self.uuid,
            properties=self.properties,
            tenant=self.tenant,
        )

    @field_validator("class_name")
    def _validate_class_name(cls, v: str) -> str:
        if len(v) == 0:
            raise ValueError("class_name must not be empty")
        return _capitalize_first_letter(v)


class Shard(BaseModel):
    """Use this class when defining a shard whose vector indexing process will be awaited for in a sync blocking fashion."""

    class_name: str
    tenant: Optional[str] = Field(default=None)

    def __hash__(self) -> int:
        return hash((self.class_name, self.tenant))


class BatchReference(BaseModel):
    """
    A reference between two objects in Weaviate.

    Performs validation on the class names and UUIDs.

    Converts provided data to an internal object containing beacons for insertion into Weaviate.
    """

    from_object_class_name: str
    from_object_uuid: UUID
    from_property_name: str
    to_object_uuid: UUID
    to_object_class_name: Optional[str] = None
    tenant: Optional[str] = None

    @field_validator("from_object_class_name")
    def _validate_from_object_class_name(cls, v: str) -> str:
        if len(v) == 0:
            raise ValueError("from_object_class_name must not be empty")
        return _capitalize_first_letter(v)

    @field_validator("to_object_class_name")
    def _validate_to_object_class_name(cls, v: Optional[str]) -> Optional[str]:
        if v is None:
            return v
        if v is not None and len(v) == 0:
            raise ValueError("to_object_class_name must not be empty if provided")
        return _capitalize_first_letter(v)

    @field_validator("to_object_uuid", "from_object_uuid")
    def _validate_uuids(cls, v: UUID) -> str:
        return get_valid_uuid(v)

    def _to_internal(self) -> _BatchReference:
        if self.to_object_class_name is None:
            self.to_object_class_name = ""
        else:
            self.to_object_class_name = self.to_object_class_name + "/"
        return _BatchReference(
            from_=f"{BEACON}{self.from_object_class_name}/{self.from_object_uuid}/{self.from_property_name}",
            to=f"{BEACON}{self.to_object_class_name}{str(self.to_object_uuid)}",
            tenant=self.tenant,
        )


@dataclass
class ErrorObject:
    """This class contains the error information for a single object in a batch operation."""

    message: str
    object_: _BatchObject
    code: Optional[int] = None
    original_uuid: Optional[UUID] = None


@dataclass
class ErrorReference:
    """This class contains the error information for a single reference in a batch operation."""

    message: str
    reference: _BatchReference
    code: Optional[int] = None


@dataclass
class BatchObjectReturn:
    """This class contains the results of a batch `insert_many` operation.

    Since the individual objects within the batch can error for differing reasons, the data is split up within this class for ease use when performing error checking, handling, and data revalidation.

    Attributes:
        `all_responses`
            A list of all the responses from the batch operation. Each response is either a `uuid_package.UUID` object or an `Error` object.
        `uuids`
            A dictionary of all the successful responses from the batch operation. The keys are the indices of the objects in the batch, and the values are the `uuid_package.UUID` objects.
        `errors`
            A dictionary of all the failed responses from the batch operation. The keys are the indices of the objects in the batch, and the values are the `Error` objects.
        `elapsed_seconds`
            The time taken to perform the batch operation.
        `has_errors`
            A boolean indicating whether or not any of the objects in the batch failed to be inserted. If this is `True`, then the `errors` dictionary will contain at least one entry.
    """

    all_responses: List[Union[uuid_package.UUID, ErrorObject]]
    uuids: Dict[int, uuid_package.UUID]
    errors: Dict[int, ErrorObject]
    elapsed_seconds: float
    has_errors: bool = False


@dataclass
class BatchReferenceReturn:
    """This class contains the results of a batch `insert_many_references` operation.

    Since the individual references within the batch can error for differing reasons, the data is split up within this class for ease use when performing error checking, handling, and data revalidation.

    Attributes:
        `elapsed_seconds`
            The time taken to perform the batch operation.
        `errors`
            A dictionary of all the failed responses from the batch operation. The keys are the indices of the references in the batch, and the values are the `Error` objects.
        `has_errors`
            A boolean indicating whether or not any of the references in the batch failed to be inserted. If this is `True`, then the `errors` dictionary will contain at least one entry.
    """

    elapsed_seconds: float
    errors: Dict[int, ErrorReference]
    has_errors: bool = False


@dataclass
class _BatchDeleteResult:
    failed: int
    matches: int
    objects: Optional[List[Dict[str, Any]]]
    successful: int



import unittest
from unittest.mock import Mock

from requests.exceptions import ConnectionError as RequestsConnectionError

from test.util import mock_connection_func, check_error_message, check_startswith_error_message
from weaviate.data.references import Reference
from weaviate.data.replication import ConsistencyLevel
from weaviate.exceptions import UnexpectedStatusCodeException


class TestReference(unittest.TestCase):
    def setUp(self):
        self.uuid_1 = "b36268d4-a6b5-5274-985f-45f13ce0c642"
        self.uuid_2 = "a36268d4-a6b5-5274-985f-45f13ce0c642"
        self.uuid_error_message = f"'uuid' must be of type str or uuid.UUID, but was: {int}"
        self.valid_uuid_error_message = "Not valid 'uuid' or 'uuid' can not be extracted from value"
        self.name_error_message = (
            lambda p: f"'from_property_name' must be of type 'str'. Given type: {p}"
        )

    def test_delete(self):
        """
        Test `delete` method`.
        """

        connection_mock = Mock()
        connection_mock.server_version = "1.13.2"
        reference = Reference(connection_mock)

        # error messages
        unexpected_error_msg = "Delete property reference to object"
        connection_error_msg = "Reference was not deleted."

        # invalid calls
        with self.assertRaises(TypeError) as error:
            reference.delete(1, "myProperty", self.uuid_2)
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(TypeError) as error:
            reference.delete(self.uuid_1, "myProperty", 2)
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(TypeError) as error:
            reference.delete(self.uuid_1, 3, self.uuid_2)
        check_error_message(self, error, self.name_error_message(int))

        with self.assertRaises(ValueError) as error:
            reference.delete("str", "myProperty", self.uuid_2)
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.delete(self.uuid_1, "myProperty", "str")
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.delete(self.uuid_1, "myProperty", self.uuid_2, consistency_level=1)

        mock_obj = mock_connection_func("delete", status_code=200)
        reference = Reference(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            reference.delete(self.uuid_1, "myProperty", self.uuid_2)
        check_startswith_error_message(self, error, unexpected_error_msg)

        mock_obj = mock_connection_func("delete", side_effect=RequestsConnectionError("Test!"))
        reference = Reference(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            reference.delete(self.uuid_1, "myProperty", self.uuid_2)
        check_error_message(self, error, connection_error_msg)

        # test valid calls
        connection_mock = mock_connection_func("delete", status_code=204)
        reference = Reference(connection_mock)

        reference.delete(self.uuid_1, "myProperty", self.uuid_2)

        connection_mock.delete.assert_called_with(
            path=f"/objects/{self.uuid_1}/references/myProperty",
            weaviate_object={"beacon": f"weaviate://localhost/{self.uuid_2}"},
            params={},
        )

        reference.delete(
            self.uuid_1,
            "hasItem",
            f"http://localhost:8080/v1/objects/{self.uuid_2}",
            consistency_level="ONE",
        )

        connection_mock.delete.assert_called_with(
            path=f"/objects/{self.uuid_1}/references/hasItem",
            weaviate_object={"beacon": f"weaviate://localhost/{self.uuid_2}"},
            params={"consistency_level": "ONE"},
        )

    def test_add(self):
        """
        Test the `add` method.
        """

        connection_mock = Mock()
        connection_mock.server_version = "1.13.2"
        reference = Reference(connection_mock)

        # error messages
        unexpected_error_msg = "Add property reference to object"
        connection_error_msg = "Reference was not added."

        # test exceptions
        with self.assertRaises(TypeError) as error:
            reference.add(1, "prop", self.uuid_1)
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(TypeError) as error:
            reference.add(self.uuid_1, 1, self.uuid_2)
        check_error_message(self, error, self.name_error_message(int))

        with self.assertRaises(TypeError) as error:
            reference.add(self.uuid_1, "prop", 1)
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.add("my UUID", "prop", self.uuid_2)
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.add(self.uuid_1, "prop", "my uuid")
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.add(self.uuid_1, "prop", self.uuid_2, consistency_level=1)

        with self.assertRaises(ValueError) as error:
            reference.add(
                f"http://localhost:8080/v1/objects/{self.uuid_1}",
                "prop",
                "http://localhost:8080/v1/objects/MY_UUID",
            )
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.add(
                "http://localhost:8080/v1/objects/My-UUID",
                "prop",
                f"http://localhost:8080/v1/objects/{self.uuid_2}",
            )
        check_error_message(self, error, self.valid_uuid_error_message)

        mock_obj = mock_connection_func("post", status_code=204)
        reference = Reference(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            reference.add(self.uuid_1, "myProperty", self.uuid_2)
        check_startswith_error_message(self, error, unexpected_error_msg)

        mock_obj = mock_connection_func("post", side_effect=RequestsConnectionError("Test!"))
        reference = Reference(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            reference.add(self.uuid_1, "myProperty", self.uuid_2)
        check_error_message(self, error, connection_error_msg)

        # test valid calls
        connection_mock = mock_connection_func("post")
        reference = Reference(connection_mock)

        # 1. Plain
        reference.add(
            "3250b0b8-eaf7-499b-ac68-9084c9c82d0f",
            "hasItem",
            "99725f35-f12a-4f36-a2e2-0d41501f4e0e",
        )
        connection_mock.post.assert_called_with(
            path="/objects/3250b0b8-eaf7-499b-ac68-9084c9c82d0f/references/hasItem",
            weaviate_object={"beacon": "weaviate://localhost/99725f35-f12a-4f36-a2e2-0d41501f4e0e"},
            params={},
        )

        # 2. using url
        reference.add(
            "http://localhost:8080/v1/objects/7591be77-5959-4386-9828-423fc5096e87",
            "hasItem",
            "http://localhost:8080/v1/objects/1cd80c11-29f0-453f-823c-21547b1511f0",
        )
        connection_mock.post.assert_called_with(
            path="/objects/7591be77-5959-4386-9828-423fc5096e87/references/hasItem",
            weaviate_object={"beacon": "weaviate://localhost/1cd80c11-29f0-453f-823c-21547b1511f0"},
            params={},
        )

        # 3. using weaviate url
        reference.add(
            "weaviate://localhost/f8def983-87e7-4e21-bf10-e32e2de3efcf",
            "hasItem",
            "weaviate://localhost/e40aaef5-d3e5-44f1-8ec4-3eafc8475078",
            consistency_level="ALL",
        )
        connection_mock.post.assert_called_with(
            path="/objects/f8def983-87e7-4e21-bf10-e32e2de3efcf/references/hasItem",
            weaviate_object={"beacon": "weaviate://localhost/e40aaef5-d3e5-44f1-8ec4-3eafc8475078"},
            params={"consistency_level": "ALL"},
        )

    def test_update(self):
        """
        Test the `update` method.
        """

        connection_mock = Mock()
        connection_mock.server_version = "1.13.2"
        reference = Reference(connection_mock)

        # error messages
        unexpected_error_msg = "Update property reference to object"
        connection_error_msg = "Reference was not updated."

        # test exceptions
        with self.assertRaises(TypeError) as error:
            reference.update(1, "prop", [self.uuid_1])
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(TypeError) as error:
            reference.update(self.uuid_1, 1, [self.uuid_2])
        check_error_message(self, error, self.name_error_message(int))

        with self.assertRaises(TypeError) as error:
            reference.update(self.uuid_1, "prop", 1)
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(TypeError) as error:
            reference.update(self.uuid_1, "prop", [1])
        check_error_message(self, error, self.uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.update("my UUID", "prop", self.uuid_2)
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.update(self.uuid_1, "prop", "my uuid")
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.update(self.uuid_1, "prop", ["my uuid"])
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.update(self.uuid_1, "prop", self.uuid_2, consistency_level=1)
        check_error_message(self, error, "1 is not a valid ConsistencyLevel")

        with self.assertRaises(ValueError) as error:
            reference.update(
                f"http://localhost:8080/v1/objects/{self.uuid_1}",
                "prop",
                "http://localhost:8080/v1/objects/MY_UUID",
            )
        check_error_message(self, error, self.valid_uuid_error_message)

        with self.assertRaises(ValueError) as error:
            reference.update(
                "http://localhost:8080/v1/objects/My-UUID",
                "prop",
                f"http://localhost:8080/v1/objects/{self.uuid_2}",
            )
        check_error_message(self, error, self.valid_uuid_error_message)

        mock_obj = mock_connection_func("put", status_code=204)
        reference = Reference(mock_obj)
        with self.assertRaises(UnexpectedStatusCodeException) as error:
            reference.update(self.uuid_1, "myProperty", self.uuid_2)
        check_startswith_error_message(self, error, unexpected_error_msg)

        mock_obj = mock_connection_func("put", side_effect=RequestsConnectionError("Test!"))
        reference = Reference(mock_obj)
        with self.assertRaises(RequestsConnectionError) as error:
            reference.update(self.uuid_1, "myProperty", self.uuid_2)
        check_error_message(self, error, connection_error_msg)

        # test valid calls
        connection_mock = mock_connection_func("put")
        reference = Reference(connection_mock)

        reference.update(
            "de998e81-fa66-440e-a1de-2a2013667e77",
            "hasAwards",
            "fc041624-4ddf-4b76-8e09-a5b0b9f9f832",
        )
        connection_mock.put.assert_called_with(
            path="/objects/de998e81-fa66-440e-a1de-2a2013667e77/references/hasAwards",
            weaviate_object=[
                {"beacon": "weaviate://localhost/fc041624-4ddf-4b76-8e09-a5b0b9f9f832"}
            ],
            params={},
        )

        reference.update(
            "4e44db9b-7f9c-4cf4-a3a0-b57024eefed0",
            "hasAwards",
            [
                "17ee17bd-a09a-49ff-adeb-d242f25f390d",
                "f8c25386-707c-40c0-b7b9-26cc0e9b2bd1",
                "d671dc52-dce4-46e7-8731-b722f19420c8",
            ],
            consistency_level=ConsistencyLevel.QUORUM,
        )
        connection_mock.put.assert_called_with(
            path="/objects/4e44db9b-7f9c-4cf4-a3a0-b57024eefed0/references/hasAwards",
            weaviate_object=[
                {"beacon": "weaviate://localhost/17ee17bd-a09a-49ff-adeb-d242f25f390d"},
                {"beacon": "weaviate://localhost/f8c25386-707c-40c0-b7b9-26cc0e9b2bd1"},
                {"beacon": "weaviate://localhost/d671dc52-dce4-46e7-8731-b722f19420c8"},
            ],
            params={"consistency_level": "QUORUM"},
        )



import json
import time
from typing import Dict

import pytest
from pytest_httpserver import HTTPServer
from werkzeug import Request, Response

import weaviate
from mock_tests.conftest import MOCK_SERVER_URL, MOCK_IP, MOCK_PORT


@pytest.mark.parametrize(
    "header",
    [
        {},
        {"Authorization": "Bearer test"},
        {"Authorization": "Bearer test", "SomethingElse": "Value"},
    ],
)
def test_additional_headers(weaviate_mock, header: Dict[str, str]):
    """Test that client sends given headers."""

    def handler(request: Request):
        assert request.headers["content-type"] == "application/json"
        for key, val in header.items():
            assert request.headers[key] == val
        return Response(json.dumps({}))

    weaviate_mock.expect_request("/v1/schema").respond_with_handler(handler)

    client = weaviate.Client(MOCK_SERVER_URL, additional_headers=header)
    client.schema.delete_all()  # some call that includes headers


@pytest.mark.parametrize("version,warning", [("1.13", True), ("1.14", False)])
def test_warning_old_weaviate(recwarn, ready_mock: HTTPServer, version: str, warning: bool):
    """Test that we warn if a new client version is using an old weaviate server."""
    ready_mock.expect_request("/v1/meta").respond_with_json({"version": version})
    weaviate.Client(MOCK_SERVER_URL)

    if warning:
        assert len(recwarn) == 2
        w = recwarn.pop()
        assert issubclass(w.category, DeprecationWarning)
        assert str(w.message).startswith("Dep001")
        w = recwarn.pop()
        assert issubclass(w.category, DeprecationWarning)
        assert str(w.message).startswith("Dep004")
    else:
        assert len(recwarn) == 1
        w = recwarn.pop()
        assert issubclass(w.category, DeprecationWarning)
        assert str(w.message).startswith("Dep004")


def test_wait_for_weaviate(httpserver: HTTPServer):
    def handler(request: Request):
        time.sleep(1)
        return Response(json.dumps({}))

    def handler_meta(request: Request):
        assert time.time() > start_time - 1
        return Response(json.dumps({"version": "1.16"}))

    httpserver.expect_request("/v1/meta").respond_with_handler(handler_meta)
    httpserver.expect_request("/v1/.well-known/ready").respond_with_handler(handler)
    start_time = time.time()
    weaviate.Client(MOCK_SERVER_URL, startup_period=30)


def test_user_pw_in_url(weaviate_mock):
    """Test that user and pw can be in the url."""
    weaviate.Client("http://user:pw@" + MOCK_IP + ":" + str(MOCK_PORT))  # no exception



"""
Error/Warning messages that are reused throughout the code.
"""


FILTER_BEACON_V14_CLS_NS_W = (
    "Based on the number of '/' in the beacon it seems that the beacon is not "
    "class namespaced. Weaviate version >= 1.14.0 STRONGLY recommends using class "
    "namespaced beacons. Non-class namespaced beacons will be removed in future "
    "versions. Class namespaced beacons look like this: "
    "'weaviate://localhost/{CLASS_NAME}/{UUID}'"
)


BATCH_MANUAL_USE_W = (
    "You are manually batching this means you are NOT using the client's "
    "built-in multi-threading. Setting `batch_size` in `client.batch.configure()` "
    "to an int value will enabled this. Also see: https://weaviate.io/developers/"
    "weaviate/current/restful-api-references/batch.html#example-request-1"
)


BATCH_REF_DEPRECATION_NEW_V14_CLS_NS_W = (
    "Weaviate Server version >= 1.14.x STRONGLY recommends using class namespaced "
    "beacons, please specify the `to_object_class_name` argument for this. The "
    "non-class namespaced beacons (None value for `to_object_class_name`) are going "
    "to be removed in the future versions of the Weaviate Server and Weaviate Python "
    "Client."
)


BATCH_REF_DEPRECATION_OLD_V14_CLS_NS_W = (
    "Weaviate Server version < 1.14.x does not support class namespaced APIs. The "
    "non-class namespaced APIs calls are going to be made instead (None value for "
    "`class_name`). The non-class namespaced APIs are going to be removed in "
    "future versions of the Weaviate Server and Weaviate Python Client. "
    "Please upgrade your Weaviate Server version."
)


DATA_DEPRECATION_NEW_V14_CLS_NS_W = (
    "Weaviate Server version >= 1.14.x STRONGLY recommends using class namespaced "
    "APIs, please specify the `class_name` argument for this. The non-class "
    "namespaced APIs (None value for `class_name`) are going to be removed in the "
    "future versions of the Weaviate Server and Weaviate Python Client."
)


DATA_DEPRECATION_OLD_V14_CLS_NS_W = (
    "Weaviate Server version < 1.14.x does not support class namespaced APIs. The "
    "non-class namespaced APIs calls are going to be made instead (None value for "
    "`class_name`). The non-class namespaced APIs are going to be removed in "
    "future versions of the Weaviate Server and Weaviate Python Client. "
    "Please upgrade your Weaviate Server version."
)


REF_DEPRECATION_NEW_V14_CLS_NS_W = (  # del
    "Weaviate Server version >= 1.14.x STRONGLY recommends using class namespaced "
    "APIs and beacons, please set the `from_class_name` AND `to_class_name` arguments "
    "for this. The non-class namespaced APIs and beacons (None value for "
    "`from_class_name` AND `to_class_name`) are going to be removed in future "
    "versions of the Weaviate Server and Weaviate Python Client."
)


REF_DEPRECATION_OLD_V14_FROM_CLS_NS_W = (
    "Weaviate Server version < 1.14.x does not support class namespaced APIs. "
    "The non-class namespaced APIs calls are going to be made instead "
    "(None value for `from_class_name`). The non-class namespaced APIs and "
    "beacons are going to be removed in future versions of the Weaviate Server "
    "and Weaviate Python Client. Please upgrade your Weaviate Server version."
)


REF_DEPRECATION_OLD_V14_TO_CLS_NS_W = (
    "Weaviate Server version < 1.14.x does not support class namespaced beacons. "
    "The non-class namespaced beacons are going to be constructed instead "
    "(None value for `to_class_name`). The non-class namespaced APIs and "
    "beacons are going to be removed in future versions of the Weaviate Server "
    "and Weaviate Python Client. Please upgrade your Weaviate Server version."
)


BATCH_EXECUTOR_SHUTDOWN_W = (
    "The BatchExecutor was shutdown, most probably when it exited the `with` statement. "
    "It will be initialized again. If you are not `batch` in the `with client.batch as batch` "
    "please make sure to shut it down when done importing data: `client.batch.shutdown()`. "
    "You can start it again using the `client.batch.start()` method."
)



import datetime
import time
from typing import Dict, Any, List

import pytest

import weaviate
from weaviate.exceptions import UnexpectedStatusCodeException, BackupFailedException

BACKUP_FILESYSTEM_PATH = "/tmp/backups"  # must be the same location as in the docker-compose file
BACKEND = "filesystem"

schema = {
    "classes": [
        {
            "class": "Paragraph",
            "properties": [
                {"dataType": ["text"], "name": "contents"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
            ],
        },
        {
            "class": "Article",
            "properties": [
                {"dataType": ["string"], "name": "title"},
                {"dataType": ["Paragraph"], "name": "hasParagraphs"},
                {"dataType": ["date"], "name": "datePublished"},
            ],
        },
    ]
}

paragraphs = [
    {"id": "fd34ccf4-1a2a-47ad-8446-231839366c3f", "properties": {"contents": "paragraph 1"}},
    {"id": "2653442b-05d8-4fa3-b46a-d4a152eb63bc", "properties": {"contents": "paragraph 2"}},
    {"id": "55374edb-17de-487f-86cb-9a9fbc30823f", "properties": {"contents": "paragraph 3"}},
    {"id": "124ff6aa-597f-44d0-8c13-62fbb1e66888", "properties": {"contents": "paragraph 4"}},
    {"id": "f787386e-7d1c-481f-b8c3-3dbfd8bbad85", "properties": {"contents": "paragraph 5"}},
]

articles = [
    {
        "id": "2fd68cbc-21ff-4e19-aaef-62531dade974",
        "properties": {
            "title": "article a",
            "datePublished": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        },
    },
    {
        "id": "7ea3f7b8-65fd-4318-a842-ae9ba38ffdca",
        "properties": {
            "title": "article b",
            "datePublished": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        },
    },
    {
        "id": "769a4280-4b85-4e67-b685-07796c49a764",
        "properties": {
            "title": "article c",
            "datePublished": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        },
    },
    {
        "id": "97fcc234-fd16-4a40-82bb-d614e9bddf8b",
        "properties": {
            "title": "article d",
            "datePublished": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        },
    },
    {
        "id": "3fa435d3-6ab2-489d-abed-c25ec526c9f4",
        "properties": {
            "title": "article e",
            "datePublished": datetime.datetime.now(datetime.timezone.utc).isoformat(),
        },
    },
]


@pytest.fixture(scope="module")
def client():
    client = weaviate.Client("http://localhost:8080")
    client.schema.create(schema)
    for para in paragraphs:
        client.data_object.create(para["properties"], "Paragraph", para["id"])
    for i, art in enumerate(articles):
        client.data_object.create(art["properties"], "Article", art["id"])
        client.data_object.reference.add(
            from_uuid=art["id"],
            from_class_name="Article",
            from_property_name="hasParagraphs",
            to_uuid=paragraphs[i]["id"],
            to_class_name="Paragraph",
        )
    yield client
    client.schema.delete_all()


def _assert_objects_exist(local_client: weaviate.Client, class_name: str, expected_count: int):
    result = local_client.query.aggregate(class_name).with_meta_count().do()
    count = result["data"]["Aggregate"][class_name][0]["meta"]["count"]
    assert (
        expected_count == count
    ), f"{class_name}: expected count: {expected_count}, received: {count}"


def _create_backup_id() -> str:
    return str(round(time.time() * 1000))


def _check_response(
    response: Dict[str, Any], backup_id: str, status: List[str], classes_include: List[str] = None
):
    assert response["id"] == backup_id
    if classes_include is not None:
        assert len(response["classes"]) == len(classes_include)
        assert sorted(response["classes"]) == sorted(classes_include)
    assert response["backend"] == "filesystem"
    assert response["path"] == f"{BACKUP_FILESYSTEM_PATH}/{backup_id}"
    assert response["status"] in status


def test_create_and_restore_backup_with_waiting(client, tmp_path):
    """Create and restore backup with waiting."""
    backup_id = _create_backup_id()
    # check data exists
    _assert_objects_exist(client, "Article", len(articles))
    _assert_objects_exist(client, "Paragraph", len(paragraphs))

    # create backup
    classes = ["Article", "Paragraph"]
    resp = client.backup.create(backup_id=backup_id, backend=BACKEND, wait_for_completion=True)
    _check_response(resp, backup_id, ["SUCCESS"], classes)

    # check data still exists
    _assert_objects_exist(client, "Article", len(articles))
    _assert_objects_exist(client, "Paragraph", len(paragraphs))

    # check create status
    resp = client.backup.get_create_status(backup_id, BACKEND)
    _check_response(resp, backup_id, ["SUCCESS"])

    # remove existing class
    client.schema.delete_class("Article")
    client.schema.delete_class("Paragraph")
    # restore backup
    resp = client.backup.restore(backup_id=backup_id, backend=BACKEND, wait_for_completion=True)
    _check_response(resp, backup_id, ["SUCCESS"], classes)

    # check data exists again
    _assert_objects_exist(client, "Article", len(articles))
    _assert_objects_exist(client, "Paragraph", len(paragraphs))
    # check restore status
    resp = client.backup.get_restore_status(backup_id, BACKEND)
    _check_response(resp, backup_id, ["SUCCESS"])


def test_create_and_restore_backup_without_waiting(client: weaviate.Client):
    """Create and restore backup without waiting."""
    backup_id = _create_backup_id()

    # check data exists
    _assert_objects_exist(client, "Article", len(articles))
    # create backup
    include = ["Article"]

    resp = client.backup.create(backup_id=backup_id, include_classes=include, backend=BACKEND)
    _check_response(resp, backup_id, ["STARTED"], include)

    # wait until created
    while True:
        resp = client.backup.get_create_status(backup_id, BACKEND)
        _check_response(resp, backup_id, ["STARTED", "TRANSFERRING", "TRANSFERRED", "SUCCESS"])
        if resp["status"] == "SUCCESS":
            break
        time.sleep(0.1)
    # check data still exists
    _assert_objects_exist(client, "Article", len(articles))
    # remove existing class
    client.schema.delete_class("Article")
    # restore backup
    resp = client.backup.restore(
        backup_id=backup_id,
        include_classes=include,
        backend=BACKEND,
    )
    _check_response(resp, backup_id, ["STARTED"], include)
    # wait until restored
    while True:
        resp = client.backup.get_restore_status(backup_id, BACKEND)
        _check_response(resp, backup_id, ["STARTED", "TRANSFERRING", "TRANSFERRED", "SUCCESS"])
        if resp["status"] == "SUCCESS":
            break
        time.sleep(0.1)
    # check data exists again
    _assert_objects_exist(client, "Article", len(articles))


def test_create_and_restore_1_of_2_classes(client: weaviate.Client):
    """Create and restore 1 of 2 classes."""
    backup_id = _create_backup_id()

    # check data exists
    _assert_objects_exist(client, "Article", len(articles))

    # create backup
    include = ["Article"]
    resp = client.backup.create(
        backup_id=backup_id, include_classes=include, backend=BACKEND, wait_for_completion=True
    )
    _check_response(resp, backup_id, ["SUCCESS"], include)

    # check data still exists
    _assert_objects_exist(client, "Article", len(articles))
    # check create status
    resp = client.backup.get_create_status(backup_id, BACKEND)
    _check_response(resp, backup_id, ["SUCCESS"])

    # remove existing class
    client.schema.delete_class("Article")
    # restore backup
    resp = client.backup.restore(
        backup_id=backup_id, include_classes=include, backend=BACKEND, wait_for_completion=True
    )
    _check_response(resp, backup_id, ["SUCCESS"], include)

    # check data exists again
    _assert_objects_exist(client, "Article", len(articles))
    # check restore status
    resp = client.backup.get_restore_status(backup_id, BACKEND)
    _check_response(resp, backup_id, ["SUCCESS"])


def test_fail_on_non_existing_backend(client: weaviate.Client):
    """Fail backup functions on non-existing backend"""
    backup_id = _create_backup_id()
    backend = "non-existing-backend"
    for func in [client.backup.create, client.backup.get_create_status, client.backup.restore]:
        with pytest.raises(ValueError) as excinfo:
            func(backup_id=backup_id, backend=backend)
            assert backend in str(excinfo.value)


def test_fail_on_non_existing_class(client: weaviate.Client):
    """Fail backup functions on non-existing class"""
    backup_id = _create_backup_id()
    class_name = "NonExistingClass"
    for func in [client.backup.create, client.backup.restore]:
        with pytest.raises(UnexpectedStatusCodeException) as excinfo:
            func(backup_id=backup_id, backend=BACKEND, include_classes=class_name)
            assert class_name in str(excinfo.value)
            assert "422" in str(excinfo.value)


def test_fail_restoring_backup_for_existing_class(client: weaviate.Client):
    """Fail restoring backup for existing class."""
    backup_id = _create_backup_id()
    class_name = ["Article"]
    resp = client.backup.create(
        backup_id=backup_id, include_classes=class_name, backend=BACKEND, wait_for_completion=True
    )
    _check_response(resp, backup_id, ["SUCCESS"], class_name)

    # fail restore
    with pytest.raises(BackupFailedException) as excinfo:
        client.backup.restore(
            backup_id=backup_id,
            include_classes=class_name,
            backend=BACKEND,
            wait_for_completion=True,
        )
        assert class_name[0] in str(excinfo.value)
        assert "already exists" in str(excinfo.value)


def test_fail_creating_existing_backup(client: weaviate.Client):
    """Fail creating existing backup"""
    backup_id = _create_backup_id()
    class_name = ["Article"]
    resp = client.backup.create(
        backup_id=backup_id, include_classes=class_name, backend=BACKEND, wait_for_completion=True
    )
    _check_response(resp, backup_id, ["SUCCESS"], class_name)

    # fail create
    with pytest.raises(UnexpectedStatusCodeException) as excinfo:
        client.backup.create(
            backup_id=backup_id,
            include_classes=class_name,
            backend=BACKEND,
            wait_for_completion=True,
        )
        assert backup_id in str(excinfo.value)
        assert "422" in str(excinfo.value)


def test_fail_restoring_non_existing_backup(client: weaviate.Client):
    """fail restoring non-existing backup"""
    backup_id = _create_backup_id()
    with pytest.raises(UnexpectedStatusCodeException) as excinfo:
        client.backup.restore(backup_id=backup_id, backend=BACKEND, wait_for_completion=True)
        assert backup_id in str(excinfo.value)
        assert "404" in str(excinfo.value)


def test_fail_checking_status_for_non_existing_restore(client: weaviate.Client):
    """Fail checking restore status for non-existing restore."""
    backup_id = _create_backup_id()
    for func in [client.backup.get_restore_status, client.backup.get_create_status]:
        with pytest.raises(UnexpectedStatusCodeException) as excinfo:
            func(
                backup_id=backup_id,
                backend=BACKEND,
            )
            assert backup_id in str(excinfo)
            assert "404" in str(excinfo)


def test_fail_creating_backup_for_both_include_and_exclude_classes(client: weaviate.Client):
    """fail creating backup for both include and exclude classes"""
    backup_id = _create_backup_id()

    for func in [client.backup.create, client.backup.restore]:
        with pytest.raises(TypeError) as excinfo:
            include = "Article"
            exclude = "Paragraph"
            func(
                backup_id=backup_id,
                include_classes=include,
                exclude_classes=exclude,
                backend=BACKEND,
                wait_for_completion=True,
            )
            assert "Either 'include_classes' OR 'exclude_classes' can be set, not both" in str(
                excinfo.value
            )



from typing import List, TypedDict


class BatchStats(TypedDict):
    queueLength: int
    ratePerSecond: int


# must use functional syntax because class is a keyword
Shard = TypedDict("Shard", {"class": str, "name": str, "objectCount": int})


class Stats(TypedDict):
    objectCount: int
    shardCount: int


class Node(TypedDict):
    batchStats: BatchStats
    gitHash: str
    name: str
    shards: List[Shard]
    stats: Stats
    status: str
    version: str



from typing import Dict, Any, List

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.collections.classes.tenants import Tenant
from weaviate.connect import Connection
from weaviate.exceptions import UnexpectedStatusCodeException


class _Tenants:
    """Represents all the CRUD methods available on a collection's multi-tenancy specification within Weaviate.

    The collection must have been created with multi-tenancy enabled in order to use any of these methods. This class
    should not be instantiated directly, but is available as a property of the `Collection` class under
    the `collection.tenants` class attribute.
    """

    def __init__(self, connection: Connection, name: str) -> None:
        self.__connection = connection
        self.__name = name

    def create(self, tenants: List[Tenant]) -> None:
        """Create the specified tenants for a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                List of tenants to add to the given collection.

        Raises:
            `requests.ConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If Weaviate reports a non-OK status.
        """
        loaded_tenants = [tenant.model_dump() for tenant in tenants]

        path = "/schema/" + self.__name + "/tenants"
        try:
            response = self.__connection.post(path=path, weaviate_object=loaded_tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                f"Collection tenants may not have been added properly for {self.__name}"
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException(
                f"Add collection tenants for {self.__name}", response
            )

    def remove(self, tenants: List[str]) -> None:
        """Remove the specified tenants from a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                List of tenant names to remove from the given class.

        Raises:
            `requests.ConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If Weaviate reports a non-OK status.
        """
        path = "/schema/" + self.__name + "/tenants"
        try:
            response = self.__connection.delete(path=path, weaviate_object=tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                f"Collection tenants may not have been deleted for {self.__name}"
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException(
                f"Delete collection tenants for {self.__name}", response
            )

    def get(self) -> Dict[str, Tenant]:
        """Return all tenants currently associated with a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Raises:
            `requests.ConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If Weaviate reports a non-OK status.
        """
        path = "/schema/" + self.__name + "/tenants"
        try:
            response = self.__connection.get(path=path)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                f"Could not get collection tenants for {self.__name}"
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException(
                f"Get collection tenants for {self.__name}", response
            )

        tenant_resp: List[Dict[str, Any]] = response.json()
        return {tenant["name"]: Tenant(**tenant) for tenant in tenant_resp}

    def update(self, tenants: List[Tenant]) -> None:
        """Update the specified tenants for a collection in Weaviate.

        The collection must have been created with multi-tenancy enabled.

        Arguments:
            `tenants`
                List of tenants to update for the given collection.

        Raises:
            `requests.ConnectionError`
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`
                If Weaviate reports a non-OK status.
        """
        loaded_tenants = [tenant.model_dump() for tenant in tenants]

        path = "/schema/" + self.__name + "/tenants"
        try:
            response = self.__connection.put(path=path, weaviate_object=loaded_tenants)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError(
                f"Collection tenants may not have been updated properly for {self.__name}"
            ) from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException(
                f"Update collection tenants for {self.__name}", response
            )



"""
GraphQL filters for `Get` and `Aggregate` commands.
GraphQL abstract class for GraphQL commands to inherit from.
"""
import warnings
from abc import ABC, abstractmethod
from copy import deepcopy
from enum import Enum
from json import dumps
from typing import Any, Tuple, Union

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.connect import Connection
from weaviate.error_msgs import FILTER_BEACON_V14_CLS_NS_W
from weaviate.util import get_vector, _sanitize_str, _decode_json_response_dict

VALUE_LIST_TYPES = {
    "valueStringList",
    "valueTextList",
    "valueIntList",
    "valueNumberList",
    "valueBooleanList",
}

VALUE_ARRAY_TYPES = {
    "valueStringArray",
    "valueTextArray",
    "valueIntArray",
    "valueNumberArray",
    "valueBooleanArray",
}

VALUE_PRIMITIVE_TYPES = {
    "valueString",
    "valueText",
    "valueInt",
    "valueNumber",
    "valueDate",
    "valueBoolean",
    "valueGeoRange",
}

ALL_VALUE_TYPES = VALUE_LIST_TYPES.union(VALUE_ARRAY_TYPES).union(VALUE_PRIMITIVE_TYPES)
VALUE_TYPES = VALUE_ARRAY_TYPES.union(VALUE_PRIMITIVE_TYPES)

WHERE_OPERATORS = [
    "And",
    "ContainsAll",
    "ContainsAny",
    "Equal",
    "GreaterThan",
    "GreaterThanEqual",
    "IsNull",
    "LessThan",
    "LessThanEqual",
    "Like",
    "NotEqual",
    "Or",
    "WithinGeoRange",
]


class MediaType(Enum):
    IMAGE = "image"
    AUDIO = "audio"
    VIDEO = "video"
    THERMAL = "thermal"
    DEPTH = "depth"
    IMU = "imu"


class GraphQL(ABC):
    """
    A base abstract class for GraphQL commands, such as Get, Aggregate.
    """

    def __init__(self, connection: Connection):
        """
        Initialize a GraphQL abstract class instance.

        Parameters
        ----------
        connection : weaviate.connect.Connection
            Connection object to an active and running weaviate instance.
        """

        self._connection = connection

    @abstractmethod
    def build(self) -> str:
        """
        Build method to be overloaded by the child classes. It should return the
        GraphQL query as a str.

        Returns
        -------
        str
            The query.
        """

    def do(self) -> dict:
        """
        Builds and runs the query.

        Returns
        -------
        dict
            The response of the query.

        Raises
        ------
        requests.ConnectionError
            If the network connection to weaviate fails.
        weaviate.UnexpectedStatusCodeException
            If weaviate reports a none OK status.
        """
        query = self.build()
        try:
            response = self._connection.post(path="/graphql", weaviate_object={"query": query})
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Query was not successful.") from conn_err

        res = _decode_json_response_dict(response, "Query was not successful")
        assert res is not None
        return res


class Filter(ABC):
    """
    A base abstract class for all filters.
    """

    def __init__(self, content: dict):
        """
        Initialize a Filter class instance.

        Parameters
        ----------
        content : dict
            The content of the `Filter` clause.
        """

        if not isinstance(content, dict):
            raise TypeError(
                f"{self.__class__.__name__} filter is expected to "
                f"be type dict but is {type(content)}"
            )
        self._content = deepcopy(content)

    @abstractmethod
    def __str__(self) -> str:
        """
        Should be implemented in each inheriting class.
        """

    @property
    def content(self) -> dict:
        return self._content


class NearText(Filter):
    """
    NearText class used to filter weaviate objects. Can be used with text models only (text2vec).
    E.g.: text2vec-contextionary, text2vec-transformers.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearText class instance.

        Parameters
        ----------
        content : dict
            The content of the `nearText` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """

        super().__init__(content)

        _check_concept(self._content)

        if "certainty" in self._content:
            if "distance" in self._content:
                raise ValueError(
                    "Cannot have both 'certainty' and 'distance' at the same time. "
                    "Only one is accepted."
                )
            _check_type(var_name="certainty", value=self._content["certainty"], dtype=float)
        if "distance" in self._content:
            _check_type(var_name="distance", value=self._content["distance"], dtype=float)

        if "moveTo" in self._content:
            _check_direction_clause(self._content["moveTo"])

        if "moveAwayFrom" in self._content:
            _check_direction_clause(self._content["moveAwayFrom"])

        if "autocorrect" in self._content:
            _check_type(var_name="autocorrect", value=self._content["autocorrect"], dtype=bool)

    def __str__(self) -> str:
        near_text = f'nearText: {{concepts: {dumps(self._content["concepts"])}'
        if "certainty" in self._content:
            near_text += f' certainty: {self._content["certainty"]}'
        if "distance" in self._content:
            near_text += f' distance: {self._content["distance"]}'
        if "moveTo" in self._content:
            move_to = self._content["moveTo"]
            near_text += f' moveTo: {{force: {move_to["force"]}'
            if "concepts" in move_to:
                near_text += f' concepts: {dumps(move_to["concepts"])}'
            if "objects" in move_to:
                near_text += _move_clause_objects_to_str(move_to["objects"])
            near_text += "}"
        if "moveAwayFrom" in self._content:
            move_away_from = self._content["moveAwayFrom"]
            near_text += f' moveAwayFrom: {{force: {move_away_from["force"]}'
            if "concepts" in move_away_from:
                near_text += f' concepts: {dumps(move_away_from["concepts"])}'
            if "objects" in move_away_from:
                near_text += _move_clause_objects_to_str(move_away_from["objects"])
            near_text += "}"
        if "autocorrect" in self._content:
            near_text += f' autocorrect: {_bool_to_str(self._content["autocorrect"])}'
        return near_text + "} "


class NearVector(Filter):
    """
    NearVector class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a NearVector class instance.

        Parameters
        ----------
        content : list
            The content of the `nearVector` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        KeyError
            If 'content' does not contain "vector".
        TypeError
            If 'content["vector"]' is not of type list.
        AttributeError
            If invalid 'content' keys are provided.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """

        super().__init__(content)

        if "vector" not in self._content:
            raise KeyError("No 'vector' key in `content` argument.")

        # Check optional fields
        if "certainty" in self._content:
            if "distance" in self._content:
                raise ValueError(
                    "Cannot have both 'certainty' and 'distance' at the same time. "
                    "Only one is accepted."
                )
            _check_type(var_name="certainty", value=self._content["certainty"], dtype=float)
        if "distance" in self._content:
            _check_type(var_name="distance", value=self._content["distance"], dtype=float)

        self._content["vector"] = get_vector(self._content["vector"])

    def __str__(self) -> str:
        near_vector = f'nearVector: {{vector: {dumps(self._content["vector"])}'
        if "certainty" in self._content:
            near_vector += f' certainty: {self._content["certainty"]}'
        if "distance" in self._content:
            near_vector += f' distance: {self._content["distance"]}'
        return near_vector + "} "


class NearObject(Filter):
    """
    NearObject class used to filter weaviate objects.
    """

    def __init__(self, content: dict, is_server_version_14: bool):
        """
        Initialize a NearVector class instance.

        Parameters
        ----------
        content : list
            The content of the `nearVector` clause.
        is_server_version_14 : bool
            Whether the Server version is >= 1.14.0.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If 'content' has key "certainty"/"distance" but the value is not float.
        TypeError
            If 'id'/'beacon' key does not have a value of type str!
        """

        super().__init__(content)

        if ("id" in self._content) and ("beacon" in self._content):
            raise ValueError("The 'content' argument should contain EITHER `id` OR `beacon`!")

        if "id" in self._content:
            self.obj_id = "id"
        else:
            self.obj_id = "beacon"
            if is_server_version_14 and len(self._content["beacon"].strip("/").split("/")) == 4:
                warnings.warn(
                    message=FILTER_BEACON_V14_CLS_NS_W,
                    category=DeprecationWarning,
                    stacklevel=1,
                )

        _check_type(var_name=self.obj_id, value=self._content[self.obj_id], dtype=str)

        if "certainty" in self._content:
            if "distance" in self._content:
                raise ValueError(
                    "Cannot have both 'certainty' and 'distance' at the same time. "
                    "Only one is accepted."
                )
            _check_type(var_name="certainty", value=self._content["certainty"], dtype=float)
        if "distance" in self._content:
            _check_type(var_name="distance", value=self._content["distance"], dtype=float)

    def __str__(self) -> str:
        near_object = f'nearObject: {{{self.obj_id}: "{self._content[self.obj_id]}"'
        if "certainty" in self._content:
            near_object += f' certainty: {self._content["certainty"]}'
        if "distance" in self._content:
            near_object += f' distance: {self._content["distance"]}'
        return near_object + "} "


class Ask(Filter):
    """
    Ask class used to filter weaviate objects by asking a question.
    """

    def __init__(self, content: dict):
        """
        Initialize a Ask class instance.

        Parameters
        ----------
        content : list
            The content of the `ask` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        TypeError
            If 'content'  has key "properties" but the type is not list or str.
        """

        super().__init__(content)

        if "question" not in self._content:
            raise ValueError('Mandatory "question" key not present in the "content"!')

        _check_type(var_name="question", value=self._content["question"], dtype=str)
        if "certainty" in self._content:
            if "distance" in self._content:
                raise ValueError(
                    "Cannot have both 'certainty' and 'distance' at the same time. "
                    "Only one is accepted."
                )
            _check_type(var_name="certainty", value=self._content["certainty"], dtype=float)
        if "distance" in self._content:
            _check_type(var_name="distance", value=self._content["distance"], dtype=float)

        if "autocorrect" in self._content:
            _check_type(var_name="autocorrect", value=self._content["autocorrect"], dtype=bool)

        if "rerank" in self._content:
            _check_type(var_name="rerank", value=self._content["rerank"], dtype=bool)

        if "properties" in self._content:
            _check_type(var_name="properties", value=self._content["properties"], dtype=(list, str))
            if isinstance(self._content["properties"], str):
                self._content["properties"] = [self._content["properties"]]

    def __str__(self) -> str:
        ask = f'ask: {{question: {dumps(self._content["question"])}'
        if "certainty" in self._content:
            ask += f' certainty: {self._content["certainty"]}'
        if "distance" in self._content:
            ask += f' distance: {self._content["distance"]}'
        if "properties" in self._content:
            ask += f' properties: {dumps(self._content["properties"])}'
        if "autocorrect" in self._content:
            ask += f' autocorrect: {_bool_to_str(self._content["autocorrect"])}'
        if "rerank" in self._content:
            ask += f' rerank: {_bool_to_str(self._content["rerank"])}'
        return ask + "} "


class NearMedia(Filter):
    def __init__(
        self,
        content: dict,
        media_type: MediaType,
    ):
        """
        Initialize a NearMedia class instance.

        Parameters
        ----------
        content : list
            The content of the `near<Media>` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["<media>"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """

        super().__init__(content)

        self._media_type = media_type

        if self._media_type.value not in self._content:
            raise ValueError(f'"content" is missing the mandatory key "{self._media_type.value}"!')

        _check_type(
            var_name=self._media_type.value, value=self._content[self._media_type.value], dtype=str
        )
        if "certainty" in self._content:
            if "distance" in self._content:
                raise ValueError(
                    "Cannot have both 'certainty' and 'distance' at the same time. "
                    "Only one is accepted."
                )
            _check_type(var_name="certainty", value=self._content["certainty"], dtype=float)
        if "distance" in self._content:
            _check_type(var_name="distance", value=self._content["distance"], dtype=float)

    def __str__(self) -> str:
        media = self._media_type.value.capitalize()
        if self._media_type == MediaType.IMU:
            media = self._media_type.value.upper()
        near_media = (
            f'near{media}: {{{self._media_type.value}: "{self._content[self._media_type.value]}"'
        )
        if "certainty" in self._content:
            near_media += f' certainty: {self._content["certainty"]}'
        if "distance" in self._content:
            near_media += f' distance: {self._content["distance"]}'
        return near_media + "} "


class NearImage(NearMedia):
    """
    NearImage class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearImage class instance.

        Parameters
        ----------
        content : list
            The content of the `nearImage` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["image"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.IMAGE)


class NearVideo(NearMedia):
    """
    NearVideo class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearVideo class instance.

        Parameters
        ----------
        content : list
            The content of the `nearVideo` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["video"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.VIDEO)


class NearAudio(NearMedia):
    """
    NearAudio class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearAudio class instance.

        Parameters
        ----------
        content : list
            The content of the `nearAudio` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["audio"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.AUDIO)


class NearDepth(NearMedia):
    """
    NearDepth class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearDepth class instance.

        Parameters
        ----------
        content : list
            The content of the `nearDepth` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["depth"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.DEPTH)


class NearThermal(NearMedia):
    """
    NearThermal class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearThermal class instance.

        Parameters
        ----------
        content : list
            The content of the `nearThermal` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["thermal"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.THERMAL)


class NearIMU(NearMedia):
    """
    NearIMU class used to filter weaviate objects.
    """

    def __init__(
        self,
        content: dict,
    ):
        """
        Initialize a NearIMU class instance.

        Parameters
        ----------
        content : list
            The content of the `nearIMU` clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        TypeError
            If 'content["imu"]' is not of type str.
        ValueError
            If 'content'  has key "certainty"/"distance" but the value is not float.
        """
        super().__init__(content, MediaType.IMU)


class Sort(Filter):
    """
    Sort filter class used to sort weaviate objects.
    """

    def __init__(self, content: Union[dict, list]):
        """
        Initialize a Where filter class instance.

        Parameters
        ----------
        content : list or dict
            The content of the `sort` filter clause or a single clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If a mandatory key is missing in the filter content.
        """

        # content is a empty list because it is going to the the list with sort clauses.

        super().__init__(content={"sort": []})

        self.add(content=content)

    def add(self, content: Union[dict, list]) -> None:
        """
        Add more sort clauses to the already existing sort clauses.

        Parameters
        ----------
        content : list or dict
            The content of the `sort` filter clause or a single clause to be added to the already
            existing ones.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If a mandatory key is missing in the filter content.
        """

        if isinstance(content, dict):
            content = [content]

        if not isinstance(content, list):
            raise TypeError(f"'content' must be of type dict or list. Given type: {type(content)}.")

        if len(content) == 0:
            raise ValueError("'content' cannot be an empty list.")

        for clause in content:
            if "path" not in clause or "order" not in clause:
                raise ValueError(
                    "One of the sort clause is missing required fields: 'path' and/or 'order'."
                )

            _check_type(
                var_name="path",
                value=clause["path"],
                dtype=list,
            )
            _check_type(
                var_name="order",
                value=clause["order"],
                dtype=str,
            )

            self._content["sort"].append(
                {
                    "path": clause["path"],
                    "order": clause["order"],
                }
            )

    def __str__(self) -> str:
        sort = "sort: ["
        for clause in self._content["sort"]:
            sort += f"{{ path: {dumps(clause['path'])} order: {clause['order']} }} "

        sort += "]"
        return sort


class Where(Filter):
    """
    Where filter class used to filter weaviate objects.
    """

    def __init__(self, content: dict):
        """
        Initialize a Where filter class instance.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        TypeError
            If 'content' is not of type dict.
        ValueError
            If a mandatory key is missing in the filter content.
        """

        super().__init__(content)

        if "path" in self._content:
            self.is_filter = True
            self._parse_filter(self._content)
        elif "operands" in self._content:
            self.is_filter = False
            self._parse_operator(self._content)
        else:
            raise ValueError(
                "Filter is missing required fields `path` or `operands`." f" Given: {self._content}"
            )

    def _parse_filter(self, content: dict) -> None:
        """
        Set filter fields for the Where filter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required field `operator`. " f"Given: {content}")
        if content["operator"] not in WHERE_OPERATORS:
            raise ValueError(
                f"Operator {content['operator']} is not allowed. "
                f"Allowed operators are: {', '.join(WHERE_OPERATORS)}"
            )
        self.path = dumps(content["path"])
        self.operator = content["operator"]
        self.value_type = _find_value_type(content)
        self.value = content[self.value_type]

        if self.operator == "WithinGeoRange" and self.value_type != "valueGeoRange":
            raise ValueError(
                f"Operator {self.operator} requires a value of type valueGeoRange. "
                f"Given value type: {self.value_type}"
            )

    def _parse_operator(self, content: dict) -> None:
        """
        Set operator fields for the Where filter.

        Parameters
        ----------
        content : dict
            The content of the `where` filter clause.

        Raises
        ------
        ValueError
            If 'content' is missing required fields.
        """

        if "operator" not in content:
            raise ValueError("Filter is missing required field `operator`." f" Given: {content}")
        if content["operator"] not in WHERE_OPERATORS:
            raise ValueError(
                f"Operator {content['operator']} is not allowed. "
                f"Allowed operators are: {WHERE_OPERATORS}"
            )
        _content = deepcopy(content)
        self.operator = _content["operator"]
        self.operands = []
        for operand in _content["operands"]:
            self.operands.append(Where(operand))

    def __str__(self) -> str:
        if self.is_filter:
            gql = f"where: {{path: {self.path} operator: {self.operator} {_convert_value_type(self.value_type)}: "
            if self.value_type in [
                "valueInt",
                "valueNumber",
                "valueIntArray",
                "valueNumberArray",
                "valueIntList",
                "valueNumberList",
            ]:
                if self.value_type in [
                    "valueIntList",
                    "valueNumberList",
                    "valueIntList",
                    "valueNumberList",
                ]:
                    _check_is_list(self.value, self.value_type)
                gql += f"{self.value}}}"
            elif self.value_type in [
                "valueText",
                "valueString",
                "valueTextList",
                "valueStringList",
                "valueTextArray",
                "valueStringArray",
            ]:
                if self.value_type in [
                    "valueTextList",
                    "valueStringList",
                    "valueTextArray",
                    "valueStringArray",
                ]:
                    _check_is_list(self.value, self.value_type)
                if isinstance(self.value, list):
                    val = [_sanitize_str(v) for v in self.value]
                    gql += f"{_render_list(val)}}}"
                else:
                    gql += f"{_sanitize_str(self.value)}}}"
            elif self.value_type in ["valueBoolean", "valueBooleanArray", "valueBooleanList"]:
                if self.value_type in ["valueBooleanArray", "valueBooleanList"]:
                    _check_is_list(self.value, self.value_type)
                if isinstance(self.value, list):
                    gql += f"{_render_list(self.value)}}}"
                else:
                    gql += f"{_bool_to_str(self.value)}}}"
            elif self.value_type == "valueGeoRange":
                _check_is_not_list(self.value, self.value_type)
                gql += f"{_geo_range_to_str(self.value)}}}"
            else:
                gql += f'"{self.value}"}}'
            return gql + " "

        operands_str = []
        for operand in self.operands:
            # remove the `where: ` from the operands and the last space
            operands_str.append(str(operand)[7:-1])
        operands = ", ".join(operands_str)
        return f"where: {{operator: {self.operator} operands: [{operands}]}} "


def _convert_value_type(_type: str) -> str:
    """Convert the value type to match `json` formatting required by the Weaviate-defined
    GraphQL endpoints. NOTE: This is crucially different to the Batch REST endpoints wherein
    the where filter is also used.

    Parameters
    ----------
    _type : str
        The Python-defined type to be converted.

    Returns
    -------
    str
        The string interpretation of the type in Weaviate-defined `json` format.
    """
    if _type == "valueTextArray" or _type == "valueTextList":
        return "valueText"
    elif _type == "valueStringArray" or _type == "valueStringList":
        return "valueString"
    elif _type == "valueIntArray" or _type == "valueIntList":
        return "valueInt"
    elif _type == "valueNumberArray" or _type == "valueNumberList":
        return "valueNumber"
    elif _type == "valueBooleanArray" or _type == "valueBooleanList":
        return "valueBoolean"
    elif _type == "valueDateArray" or _type == "valueDateList":
        return "valueDate"
    else:
        return _type


def _render_list(value: list) -> str:
    """Convert a list of values to string (lowercased) to match `json` formatting.

    Parameters
    ----------
    value : list
        The value to be converted

    Returns
    -------
    str
        The string interpretation of the value in `json` format.
    """
    return f'[{",".join(value)}]'


def _check_is_list(value: Any, _type: str) -> None:
    """Checks whether the provided value is a list to match the given `value_type`.

    Parameters
    ----------
    value : list
        The value to be checked.
    _type : str
        The type to be checked against.

    Raises
    ------
    TypeError
        If the value is not a list.
    """
    if not isinstance(value, list):
        raise TypeError(
            f"Must provide a list when constructing where filter for {_type} with {value}"
        )


def _check_is_not_list(value: Any, _type: str) -> None:
    """Checks whether the provided value is a list to match the given `value_type`.

    Parameters
    ----------
    value : list
        The value to be checked.
    _type : str
        The type to be checked against.

    Raises
    ------
    TypeError
        If the value is a list.
    """
    if isinstance(value, list):
        raise TypeError(
            f"Cannot provide a list when constructing where filter for {_type} with {value}"
        )


def _geo_range_to_str(value: dict) -> str:
    """
    Convert the valueGeoRange object to match `json` formatting.

    Parameters
    ----------
    value : dict
        The value to be converted.

    Returns
    -------
    str
        The string interpretation of the value in `json` format.
    """
    latitude = value["geoCoordinates"]["latitude"]
    longitude = value["geoCoordinates"]["longitude"]
    distance = value["distance"]["max"]
    return f"{{ geoCoordinates: {{ latitude: {latitude} longitude: {longitude} }} distance: {{ max: {distance} }}}}"


def _bool_to_str(value: bool) -> str:
    """
    Convert a bool value to string (lowercased) to match `json` formatting.

    Parameters
    ----------
    value : bool
        The value to be converted

    Returns
    -------
    str
        The string interpretation of the value in `json` format.
    """

    if value is True:
        return "true"
    return "false"


def _check_direction_clause(direction: dict) -> None:
    """
    Validate the direction sub clause.

    Parameters
    ----------
    direction : dict
        A sub clause of the Explore filter.

    Raises
    ------
    TypeError
        If 'direction' is not a dict.
    TypeError
        If the value of the "force" key is not float.
    ValueError
        If no "force" key in the 'direction'.
    """

    _check_type(var_name="moveXXX", value=direction, dtype=dict)

    if ("concepts" not in direction) and ("objects" not in direction):
        raise ValueError("The 'move' clause should contain `concepts` OR/AND `objects`!")

    if "concepts" in direction:
        _check_concept(direction)
    if "objects" in direction:
        _check_objects(direction)
    if "force" not in direction:
        raise ValueError("'move' clause needs to state a 'force'")
    _check_type(var_name="force", value=direction["force"], dtype=float)


def _check_concept(content: dict) -> None:
    """
    Validate the concept sub clause.

    Parameters
    ----------
    content : dict
        An Explore (sub) clause to check for 'concepts'.

    Raises
    ------
    ValueError
        If no "concepts" key in the 'content' dict.
    TypeError
        If the value of the  "concepts" is of wrong type.
    """

    if "concepts" not in content:
        raise ValueError("No concepts in content")

    _check_type(
        var_name="concepts",
        value=content["concepts"],
        dtype=(list, str),
    )
    if isinstance(content["concepts"], str):
        content["concepts"] = [content["concepts"]]


def _check_objects(content: dict) -> None:
    """
    Validate the `objects` sub clause of the `move` clause.

    Parameters
    ----------
    content : dict
        An Explore (sub) clause to check for 'objects'.

    Raises
    ------
    ValueError
        If no "concepts" key in the 'content' dict.
    TypeError
        If the value of the  "concepts" is of wrong type.
    """

    _check_type(var_name="objects", value=content["objects"], dtype=(list, dict))
    if isinstance(content["objects"], dict):
        content["objects"] = [content["objects"]]

    if len(content["objects"]) == 0:
        raise ValueError("'moveXXX' clause specifies 'objects' but no value provided.")

    for obj in content["objects"]:
        if len(obj) != 1 or ("id" not in obj and "beacon" not in obj):
            raise ValueError(
                "Each object from the `move` clause should have ONLY `id` OR " "`beacon`!"
            )


def _check_type(var_name: str, value: Any, dtype: Union[Tuple[type, type], type]) -> None:
    """
    Check key-value type.

    Parameters
    ----------
    var_name : str
        The variable name for which to check the type (used for error message)!
    value : Any
        The value for which to check the type.
    dtype : Union[Tuple[type, type], type]
        The expected data type of the `value`.

    Raises
    ------
    TypeError
        If the `value` type does not match the expected `dtype`.
    """

    if not isinstance(value, dtype):
        raise TypeError(
            f"'{var_name}' key-value is expected to be of type {dtype} but is {type(value)}!"
        )


def _find_value_type(content: dict) -> str:
    """
    Find the correct type of the content.

    Parameters
    ----------
    content : dict
        The content for which to find the appropriate data type.

    Returns
    -------
    str
        The correct data type.

    Raises
    ------
    ValueError
        If missing required fields.
    """

    value_type = ALL_VALUE_TYPES & set(content.keys())

    if len(value_type) == 0:
        raise ValueError(
            f"'value<TYPE>' field is either missing or incorrect: {content}. Valid values are: {VALUE_TYPES}."
        )
    if len(value_type) != 1:
        raise ValueError(f"Multiple fields 'value<TYPE>' are not supported: {content}")

    return value_type.pop()


def _move_clause_objects_to_str(objects: list) -> str:
    """
    _summary_

    Parameters
    ----------
    objects : list
        _description_

    Returns
    -------
    str
        _description_
    """

    to_return = " objects: ["
    for obj in objects:
        if "id" in obj:
            id_beacon = "id"
        else:
            id_beacon = "beacon"
        to_return += f"{{{id_beacon}: {dumps(obj[id_beacon])}}} "

    return to_return + "]"



from dataclasses import dataclass, field
from typing import Optional, Tuple, Union

from pydantic import BaseModel, Field


@dataclass
class ConnectionConfig:
    session_pool_connections: int = 20
    session_pool_maxsize: int = 20

    def __post_init__(self) -> None:
        if not isinstance(self.session_pool_connections, int):
            raise TypeError(
                f"session_pool_connections must be {int}, received {type(self.session_pool_connections)}"
            )
        if not isinstance(self.session_pool_maxsize, int):
            raise TypeError(
                f"session_pool_maxsize must be {int}, received {type(self.session_pool_maxsize)}"
            )


@dataclass
class Config:
    grpc_port_experimental: Optional[int] = None
    grpc_secure_experimental: bool = False
    connection_config: ConnectionConfig = field(default_factory=ConnectionConfig)

    def __post_init__(self) -> None:
        if self.grpc_port_experimental is not None and not isinstance(
            self.grpc_port_experimental, int
        ):
            raise TypeError(
                f"grpc_port_experimental must be {int}, received {type(self.grpc_port_experimental)}"
            )
        if not isinstance(self.grpc_secure_experimental, bool):
            raise TypeError(
                f"grpc_secure_experimental must be {bool}, received {type(self.grpc_secure_experimental)}"
            )


class AdditionalConfig(BaseModel):
    connection: ConnectionConfig = Field(default_factory=ConnectionConfig)
    proxies: Union[dict, str, None] = Field(default=None)
    startup_period: int = Field(default=5)
    timeout: Tuple[int, int] = Field(default_factory=lambda: (10, 60))
    trust_env: bool = Field(default=False)



import pytest

import weaviate
from weaviate.collections.classes.config import (
    Configure,
    Property,
    ConsistencyLevel,
    DataType,
)
from weaviate.collections.classes.data import DataObject
from weaviate.collections.classes.grpc import MetadataQuery


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local(port=8087, grpc_port=50058)
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


@pytest.mark.parametrize(
    "level", [ConsistencyLevel.ONE, ConsistencyLevel.ALL, ConsistencyLevel.QUORUM]
)
def test_consistency_on_multinode(client: weaviate.WeaviateClient, level: ConsistencyLevel):
    name = "TestConsistency"
    client.collections.delete(name)
    collection = client.collections.create(
        name=name,
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="name", data_type=DataType.TEXT),
        ],
        replication_config=Configure.replication(factor=2),
    ).with_consistency_level(level)

    collection.data.insert({"name": "first"})
    ret = collection.data.insert_many(
        [DataObject(properties={"name": "second"}), DataObject(properties={"name": "third"})]
    )
    assert not ret.has_errors

    objects = collection.query.fetch_objects(
        return_metadata=MetadataQuery(is_consistent=True)
    ).objects
    for obj in objects:
        assert obj.metadata.is_consistent



import datetime
import uuid as uuid_package
from typing import (
    Dict,
    Any,
    Optional,
    List,
    Tuple,
    Generic,
    Type,
    Union,
    cast,
    get_type_hints,
    get_origin,
)

from requests.exceptions import ConnectionError as RequestsConnectionError

from weaviate.collections.classes.batch import (
    _BatchObject,
    BatchObjectReturn,
    _BatchReference,
    BatchReferenceReturn,
    _BatchDeleteResult,
)
from weaviate.collections.classes.config import ConsistencyLevel
from weaviate.collections.classes.data import (
    DataObject,
    DataReference,
)
from weaviate.collections.classes.internal import (
    _Object,
    _metadata_from_dict,
    _Reference,
)
from weaviate.collections.classes.orm import (
    Model,
)
from weaviate.collections.classes.types import Properties, TProperties, _check_data_model
from weaviate.collections.classes.filters import _Filters
from weaviate.collections.batch.grpc import _BatchGRPC, _validate_props
from weaviate.collections.batch.rest import _BatchREST
from weaviate.connect import Connection
from weaviate.exceptions import (
    UnexpectedStatusCodeException,
    ObjectAlreadyExistsException,
)
from weaviate.util import (
    _datetime_to_string,
    _decode_json_response_dict,
)
from weaviate.types import BEACON, UUID


class _Data:
    def __init__(
        self,
        connection: Connection,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
    ) -> None:
        self._connection = connection
        self.name = name
        self._consistency_level = consistency_level
        self._tenant = tenant
        self._batch_grpc = _BatchGRPC(connection, consistency_level)
        self._batch_rest = _BatchREST(connection, consistency_level)

    def _insert(self, weaviate_obj: Dict[str, Any], clean_props: bool) -> uuid_package.UUID:
        path = "/objects"
        _validate_props(weaviate_obj["properties"], clean_props=clean_props)

        params, weaviate_obj = self.__apply_context_to_params_and_object({}, weaviate_obj)
        try:
            response = self._connection.post(path=path, weaviate_object=weaviate_obj, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not added to Weaviate.") from conn_err
        if response.status_code == 200:
            return uuid_package.UUID(weaviate_obj["id"])

        try:
            response_json = _decode_json_response_dict(response, "insert object")
            assert response_json is not None
            if "already exists" in response_json["error"][0]["message"]:
                raise ObjectAlreadyExistsException(weaviate_obj["id"])
        except KeyError:
            pass
        raise UnexpectedStatusCodeException("Creating object", response)

    def delete_by_id(self, uuid: UUID) -> bool:
        """Delete an object from the collection based on its UUID.

        Arguments:
            `uuid`
                The UUID of the object to delete, REQUIRED.
        """
        path = f"/objects/{self.name}/{uuid}"

        try:
            response = self._connection.delete(path=path, params=self.__apply_context({}))
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object could not be deleted.") from conn_err
        if response.status_code == 204:
            return True  # Successfully deleted
        elif response.status_code == 404:
            return False  # did not exist
        raise UnexpectedStatusCodeException("Delete object", response)

    def delete_many(
        self, where: _Filters, verbose: bool = False, dry_run: bool = False
    ) -> _BatchDeleteResult:
        """Delete multiple objects from the collection based on a filter.

        Arguments:
            `where`
                The filter to apply. This filter is the same that is used when performing queries and has the same syntax, REQUIRED.
            `verbose`
                Whether to return the deleted objects in the response.
            `dry_run`
                Whether to perform a dry run. If set to `True`, the objects will not be deleted, but the response will contain the objects that would have been deleted.

        Raises:
            `requests.ConnectionError`:
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`:
                If Weaviate reports a non-OK status.
        """
        return self._batch_rest.delete(self.name, where, verbose, dry_run, self._tenant)

    def _replace(self, weaviate_obj: Dict[str, Any], uuid: UUID) -> None:
        path = f"/objects/{self.name}/{uuid}"
        params, weaviate_obj = self.__apply_context_to_params_and_object({}, weaviate_obj)

        weaviate_obj["id"] = str(uuid)  # must add ID to payload for PUT request

        try:
            response = self._connection.put(path=path, weaviate_object=weaviate_obj, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not replaced.") from conn_err
        if response.status_code == 200:
            return
        raise UnexpectedStatusCodeException("Replacing object", response)

    def _update(self, weaviate_obj: Dict[str, Any], uuid: UUID) -> None:
        path = f"/objects/{self.name}/{uuid}"
        params, weaviate_obj = self.__apply_context_to_params_and_object({}, weaviate_obj)

        try:
            response = self._connection.patch(
                path=path, weaviate_object=weaviate_obj, params=params
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Object was not updated.") from conn_err
        if response.status_code == 204:
            return
        raise UnexpectedStatusCodeException("Update object", response)

    def _get_by_id(self, uuid: UUID, include_vector: bool) -> Optional[Dict[str, Any]]:
        path = f"/objects/{self.name}/{uuid}"
        params: Dict[str, Any] = {}
        if include_vector:
            params["include"] = "vector"
        return self._get_from_weaviate(params=self.__apply_context(params), path=path)

    def _get(self, limit: Optional[int], include_vector: bool) -> Optional[Dict[str, Any]]:
        path = "/objects"
        params: Dict[str, Any] = {"class": self.name}
        if limit is not None:
            params["limit"] = limit
        if include_vector:
            params["include"] = "vector"
        return self._get_from_weaviate(params=self.__apply_context(params), path=path)

    def _get_from_weaviate(self, params: Dict[str, Any], path: str) -> Optional[Dict[str, Any]]:
        try:
            response = self._connection.get(path=path, params=params)
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Could not get object/s.") from conn_err
        if response.status_code == 200:
            response_json = _decode_json_response_dict(response, "get")
            assert response_json is not None
            return response_json
        if response.status_code == 404:
            return None
        raise UnexpectedStatusCodeException("Get object/s", response)

    def _reference_add(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        params: Dict[str, str] = {}

        path = f"/objects/{self.name}/{from_uuid}/references/{from_property}"
        for beacon in ref._to_beacons():
            try:
                response = self._connection.post(
                    path=path,
                    weaviate_object=beacon,
                    params=self.__apply_context(params),
                )
            except RequestsConnectionError as conn_err:
                raise RequestsConnectionError("Reference was not added.") from conn_err
            if response.status_code != 200:
                raise UnexpectedStatusCodeException("Add property reference to object", response)

    def _reference_add_many(self, refs: List[DataReference]) -> BatchReferenceReturn:
        return self._batch_rest.references(
            [
                _BatchReference(
                    from_=f"{BEACON}{self.name}/{ref.from_uuid}/{ref.from_property}",
                    to=f"{BEACON}{ref.to_uuid}",
                    tenant=self._tenant,
                )
                for ref in refs
            ]
        )

    def _reference_delete(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        params: Dict[str, str] = {}

        path = f"/objects/{self.name}/{from_uuid}/references/{from_property}"
        for beacon in ref._to_beacons():
            try:
                response = self._connection.delete(
                    path=path,
                    weaviate_object=beacon,
                    params=self.__apply_context(params),
                )
            except RequestsConnectionError as conn_err:
                raise RequestsConnectionError("Reference was not added.") from conn_err
            if response.status_code != 204:
                raise UnexpectedStatusCodeException("Add property reference to object", response)

    def _reference_replace(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        params: Dict[str, str] = {}

        path = f"/objects/{self.name}/{from_uuid}/references/{from_property}"
        try:
            response = self._connection.put(
                path=path,
                weaviate_object=ref._to_beacons(),
                params=self.__apply_context(params),
            )
        except RequestsConnectionError as conn_err:
            raise RequestsConnectionError("Reference was not added.") from conn_err
        if response.status_code != 200:
            raise UnexpectedStatusCodeException("Add property reference to object", response)

    def __apply_context(self, params: Dict[str, Any]) -> Dict[str, Any]:
        if self._tenant is not None:
            params["tenant"] = self._tenant
        if self._consistency_level is not None:
            params["consistency_level"] = self._consistency_level
        return params

    def __apply_context_to_params_and_object(
        self, params: Dict[str, Any], obj: Dict[str, Any]
    ) -> Tuple[Dict[str, Any], Dict[str, Any]]:
        if self._tenant is not None:
            obj["tenant"] = self._tenant
        if self._consistency_level is not None:
            params["consistency_level"] = self._consistency_level
        return params, obj

    def _serialize_properties(self, data: Properties) -> Dict[str, Any]:
        return {
            key: val._to_beacons()
            if isinstance(val, _Reference)
            else self.__serialize_primitive(val)
            for key, val in data.items()
        }

    def __serialize_primitive(self, value: Any) -> Any:
        if isinstance(value, uuid_package.UUID):
            return str(value)
        if isinstance(value, datetime.datetime):
            return _datetime_to_string(value)
        if isinstance(value, list):
            return [self.__serialize_primitive(val) for val in value]
        return value

    def _deserialize_primitive(self, value: Any, type_value: Optional[Any]) -> Any:
        if type_value is None:
            return value
        if type_value == uuid_package.UUID:
            return uuid_package.UUID(value)
        if type_value == datetime.datetime:
            return datetime.datetime.fromisoformat(value)
        if isinstance(type_value, list):
            return [
                self._deserialize_primitive(val, type_value[idx]) for idx, val in enumerate(value)
            ]
        return value


class _DataCollection(Generic[Properties], _Data):
    def __init__(
        self,
        connection: Connection,
        name: str,
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
        type_: Optional[Type[Properties]] = None,
    ):
        super().__init__(connection, name, consistency_level, tenant)
        self.__type = type_

    def with_data_model(self, data_model: Type[TProperties]) -> "_DataCollection[TProperties]":
        _check_data_model(data_model)
        return _DataCollection[TProperties](
            self._connection, self.name, self._consistency_level, self._tenant, data_model
        )

    def __deserialize_properties(self, data: Dict[str, Any]) -> Properties:
        hints = (
            get_type_hints(self.__type)
            if self.__type and not get_origin(self.__type) == dict
            else {}
        )
        return cast(
            Properties,
            {key: self._deserialize_primitive(val, hints.get(key)) for key, val in data.items()},
        )

    def _json_to_object(self, obj: Dict[str, Any]) -> _Object[Properties]:
        props = self.__deserialize_properties(obj["properties"])
        return _Object(
            properties=cast(Properties, props),
            metadata=_metadata_from_dict(obj),
        )

    def insert(
        self,
        properties: Properties,
        uuid: Optional[UUID] = None,
        vector: Optional[List[float]] = None,
    ) -> uuid_package.UUID:
        """Insert a single object into the collection.

        Arguments:
            `properties`
                The properties of the object, REQUIRED.
            `uuid`
                The UUID of the object. If not provided, a random UUID will be generated.
            `vector`
                The vector of the object.
        """
        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(properties),
            "id": str(uuid if uuid is not None else uuid_package.uuid4()),
        }

        if vector is not None:
            weaviate_obj["vector"] = vector

        return self._insert(weaviate_obj, False)

    def insert_many(
        self,
        objects: List[Union[Properties, DataObject[Properties]]],
    ) -> BatchObjectReturn:
        """Insert multiple objects into the collection.

        Arguments:
            `objects`
                The objects to insert. This can be either a list of `Properties` or `DataObject[Properties]`
                    If you didn't set `data_model` then `Properties` will be `Data[str, Any]` in which case you can insert simple dictionaries here.
                        If you want to insert vectors and UUIDs alongside your properties, you will have to use `DataObject` instead.

        Raises:
            `weaviate.exceptions.WeaviateQueryException`:
                If the network connection to Weaviate fails.
            `weaviate.exceptions.WeaviateInsertInvalidPropertyError`:
                If a property is invalid. I.e., has name `id` or `vector`, which are reserved.
        """
        return self._batch_grpc.objects(
            [
                _BatchObject(
                    class_name=self.name,
                    vector=obj.vector,
                    uuid=obj.uuid,
                    properties=obj.properties,
                    tenant=self._tenant,
                )
                if isinstance(obj, DataObject) and isinstance(obj.properties, dict)
                else _BatchObject(
                    class_name=self.name,
                    vector=None,
                    uuid=None,
                    properties=cast(dict, obj),
                    tenant=None,
                )
                for obj in objects
            ]
        )

    def replace(
        self, properties: Properties, uuid: UUID, vector: Optional[List[float]] = None
    ) -> None:
        """Replace an object in the collection.

        This is equivalent to a PUT operation.

        Arguments:
            `properties`
                The properties of the object, REQUIRED.
            `uuid`
                The UUID of the object, REQUIRED.
            `vector`
                The vector of the object.

        Raises:
            `requests.ConnectionError`:
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`:
                If Weaviate reports a non-OK status.
            `weaviate.exceptions.WeaviateInsertInvalidPropertyError`:
                If a property is invalid. I.e., has name `id` or `vector`, which are reserved.
        """
        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(properties),
        }
        if vector is not None:
            weaviate_obj["vector"] = vector

        self._replace(weaviate_obj, uuid=uuid)

    def update(
        self, properties: Properties, uuid: UUID, vector: Optional[List[float]] = None
    ) -> None:
        """Update an object in the collection.

        This is equivalent to a PATCH operation.

        Arguments:
            `properties`
                The properties of the object, REQUIRED.
            `uuid`
                The UUID of the object, REQUIRED.
            `vector`
                The vector of the object.
        """
        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(properties),
        }
        if vector is not None:
            weaviate_obj["vector"] = vector

        self._update(weaviate_obj, uuid=uuid)

    def reference_add(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        """Create a reference between an object in this collection and any other object in Weaviate.

        Arguments:
            `from_uuid`
                The UUID of the object in this collection, REQUIRED.
            `from_property`
                The name of the property in the object in this collection, REQUIRED.
            `ref`
                The reference to add, REQUIRED. Use `Reference.to` to generate the correct type.

        Raises:
            `requests.ConnectionError`:
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException`:
                If Weaviate reports a non-OK status.
        """
        self._reference_add(
            from_uuid=from_uuid,
            from_property=from_property,
            ref=ref,
        )

    def reference_add_many(self, refs: List[DataReference]) -> BatchReferenceReturn:
        """Create multiple references on a property in batch between objects in this collection and any other object in Weaviate.

        Arguments:
            `refs`
                The references to add including the prop name, from UUID, and to UUID.

        Returns:
            `BatchReferenceReturn`
                A `BatchReferenceReturn` object containing the results of the batch operation.

        Raises:
            `requests.ConnectionError`:
                If the network connection to Weaviate fails.
            `weaviate.UnexpectedStatusCodeException
                If Weaviate reports a non-OK status.
        """
        return self._reference_add_many(refs)

    def reference_delete(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        """Delete a reference from an object within the collection.

        Arguments:
            `from_uuid`
                The UUID of the object in this collection, REQUIRED.
            `from_property`
                The name of the property in the object in this collection from which the reference should be deleted, REQUIRED.
            `ref`
                The reference to delete, REQUIRED. Use `Reference.to` to generate the correct type.
        """
        self._reference_delete(from_uuid=from_uuid, from_property=from_property, ref=ref)

    def reference_replace(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        """Replace a reference of an object within the collection.

        Arguments:
            `from_uuid`
                The UUID of the object in this collection, REQUIRED.
            `from_property`
                The name of the property in the object in this collection from which the reference should be replaced, REQUIRED.
            `ref`
                The reference to replace, REQUIRED. Use `Reference.to` to generate the correct type.
        """
        self._reference_replace(from_uuid=from_uuid, from_property=from_property, ref=ref)


class _DataCollectionModel(Generic[Model], _Data):
    def __init__(
        self,
        connection: Connection,
        name: str,
        model: Type[Model],
        consistency_level: Optional[ConsistencyLevel],
        tenant: Optional[str],
    ):
        super().__init__(connection, name, consistency_level, tenant)
        self.__model = model

    def _json_to_object(self, obj: Dict[str, Any]) -> _Object[Model]:
        for ref in self.__model.get_ref_fields(self.__model):
            if ref not in obj["properties"]:
                continue

            beacons = obj["properties"][ref]
            uuids = []
            for beacon in beacons:
                uri = beacon["beacon"]
                assert isinstance(uri, str)
                uuids.append(uri.split("/")[-1])

            obj["properties"][ref] = uuids

        # weaviate does not save none values, so we need to add them to pass model validation
        for prop in self.__model.get_non_default_fields(self.__model):
            if prop not in obj["properties"]:
                obj["properties"][prop] = None

        metadata = _metadata_from_dict(obj)
        model_object = _Object[Model](
            properties=self.__model.model_validate(
                {
                    **obj["properties"],
                    "uuid": metadata.uuid,
                    "vector": metadata.vector,
                }
            ),
            metadata=metadata,
        )
        return model_object

    def insert(self, obj: Model) -> uuid_package.UUID:
        self.__model.model_validate(obj)
        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(obj.props_to_dict()),
            "id": str(obj.uuid),
        }
        if obj.vector is not None:
            weaviate_obj["vector"] = obj.vector

        self._insert(weaviate_obj, False)
        return uuid_package.UUID(str(obj.uuid))

    def insert_many(self, objects: List[Model]) -> BatchObjectReturn:
        for obj in objects:
            self.__model.model_validate(obj)

        data_objects = [
            _BatchObject(
                class_name=self.name,
                properties=obj.props_to_dict(),
                tenant=self._tenant,
                uuid=obj.uuid,
                vector=obj.vector,
            )
            for obj in objects
        ]

        return self._batch_grpc.objects(data_objects)

    def replace(self, obj: Model, uuid: UUID) -> None:
        self.__model.model_validate(obj)

        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(obj.props_to_dict()),
        }
        if obj.vector is not None:
            weaviate_obj["vector"] = obj.vector

        self._replace(weaviate_obj, uuid)

    def update(self, obj: Model, uuid: UUID) -> None:
        self.__model.model_validate(obj)

        weaviate_obj: Dict[str, Any] = {
            "class": self.name,
            "properties": self._serialize_properties(obj.props_to_dict()),
        }
        if obj.vector is not None:
            weaviate_obj["vector"] = obj.vector

        self._update(weaviate_obj, uuid)

    def get_by_id(self, uuid: UUID, include_vector: bool = False) -> Optional[_Object[Model]]:
        ret = self._get_by_id(uuid=uuid, include_vector=include_vector)
        if ret is None:
            return None
        return self._json_to_object(ret)

    def get(
        self, limit: Optional[int] = None, include_vector: bool = False
    ) -> List[_Object[Model]]:
        ret = self._get(limit=limit, include_vector=include_vector)
        if ret is None:
            return []

        return [self._json_to_object(obj) for obj in ret["objects"]]

    def reference_add(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        self._reference_add(from_uuid=from_uuid, from_property=from_property, ref=ref)

    def reference_delete(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        self._reference_delete(from_uuid=from_uuid, from_property=from_property, ref=ref)

    def reference_replace(self, from_uuid: UUID, from_property: str, ref: _Reference) -> None:
        self._reference_replace(from_uuid=from_uuid, from_property=from_property, ref=ref)

    def reference_add_many(self, refs: List[DataReference]) -> BatchReferenceReturn:
        return self._reference_add_many(refs)



import sys
from typing import List, Optional

if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated

import pytest as pytest

from weaviate.collections.classes.config import PropertyConfig
from weaviate.collections.classes.orm import BaseProperty, Reference
from weaviate.types import UUIDS


class Group(BaseProperty):
    name: str


@pytest.mark.parametrize(
    "member_type,expected",
    [
        (str, "text"),
        (int, "int"),
        (float, "number"),
        (List[str], "text[]"),
        (List[int], "int[]"),
        (List[float], "number[]"),
    ],
)
@pytest.mark.parametrize("optional", [True, False])
def test_types(member_type, expected: str, optional: bool):
    if optional:
        member_type = Optional[member_type]

    class ModelTypes(BaseProperty):
        name: member_type

    non_optional_types = ModelTypes.type_to_dict(ModelTypes)
    assert non_optional_types[0]["dataType"][0] == expected


@pytest.mark.parametrize(
    "member_type, annotation ,expected",
    [
        (str, PropertyConfig(index_filterable=False), "text"),
        (UUIDS, Reference(Group), "Group"),
        (Optional[UUIDS], Reference(Group), "Group"),
    ],
)
def test_types_annotation(member_type, annotation, expected: str):
    member_type = Annotated[member_type, annotation]

    class ModelTypes(BaseProperty):
        name: member_type

    non_optional_types = ModelTypes.type_to_dict(ModelTypes)
    assert non_optional_types[0]["dataType"][0] == expected



import sys
from typing import TypedDict

import pytest as pytest
import uuid

from weaviate.collections.classes.data import DataObject, DataReference
from weaviate.collections.classes.grpc import FromReference, FromReferenceMultiTarget, MetadataQuery


if sys.version_info < (3, 9):
    from typing_extensions import Annotated
else:
    from typing import Annotated


import weaviate
from weaviate.collections.classes.config import (
    Configure,
    Property,
    DataType,
    ReferenceProperty,
    ReferencePropertyMultiTarget,
)

from weaviate.collections.classes.internal import CrossReference, Reference


@pytest.fixture(scope="module")
def client():
    client = weaviate.connect_to_local()
    client.collections.delete_all()
    yield client
    client.collections.delete_all()


def test_reference_add_delete_replace(client: weaviate.WeaviateClient):
    ref_collection = client.collections.create(
        name="RefClass2", vectorizer_config=Configure.Vectorizer.none()
    )
    uuid_to = ref_collection.data.insert(properties={})
    collection = client.collections.create(
        name="SomethingElse",
        properties=[ReferenceProperty(name="ref", target_collection="RefClass2")],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    uuid_from1 = collection.data.insert({}, uuid.uuid4())
    uuid_from2 = collection.data.insert({"ref": Reference.to(uuids=uuid_to)}, uuid.uuid4())
    collection.data.reference_add(
        from_uuid=uuid_from1, from_property="ref", ref=Reference.to(uuids=uuid_to)
    )

    collection.data.reference_delete(
        from_uuid=uuid_from1, from_property="ref", ref=Reference.to(uuids=uuid_to)
    )
    assert len(collection.query.fetch_object_by_id(uuid_from1).properties["ref"]) == 0

    collection.data.reference_add(
        from_uuid=uuid_from2, from_property="ref", ref=Reference.to(uuids=uuid_to)
    )
    obj = collection.query.fetch_object_by_id(uuid_from2)
    assert len(obj.properties["ref"]) == 2
    assert str(uuid_to) in "".join([ref["beacon"] for ref in obj.properties["ref"]])

    collection.data.reference_replace(
        from_uuid=uuid_from2, from_property="ref", ref=Reference.to(uuids=[])
    )
    assert len(collection.query.fetch_object_by_id(uuid_from2).properties["ref"]) == 0

    client.collections.delete("SomethingElse")
    client.collections.delete("RefClass2")


def test_mono_references_grpc(client: weaviate.WeaviateClient):
    A = client.collections.create(
        name="A",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    uuid_A1 = A.data.insert(properties={"Name": "A1"})
    uuid_A2 = A.data.insert(properties={"Name": "A2"})

    objects = A.query.bm25(query="A1", return_properties="name").objects
    assert objects[0].properties["name"] == "A1"

    B = client.collections.create(
        name="B",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            ReferenceProperty(name="ref", target_collection="A"),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_B = B.data.insert({"Name": "B", "ref": Reference.to(uuids=uuid_A1)})
    B.data.reference_add(from_uuid=uuid_B, from_property="ref", ref=Reference.to(uuids=uuid_A2))

    objects = B.query.bm25(
        query="B",
        return_properties=FromReference(
            link_on="ref",
            return_properties=["name"],
        ),
    ).objects
    assert objects[0].properties["ref"].objects[0].properties["name"] == "A1"
    assert objects[0].properties["ref"].objects[1].properties["name"] == "A2"

    objects = B.query.bm25(
        query="B",
        return_properties=[
            FromReference(
                link_on="ref",
                return_properties=["name"],
                return_metadata=MetadataQuery(uuid=True),
            )
        ],
    ).objects
    assert objects[0].properties["ref"].objects[0].properties["name"] == "A1"
    assert objects[0].properties["ref"].objects[0].metadata.uuid == uuid_A1
    assert objects[0].properties["ref"].objects[1].properties["name"] == "A2"
    assert objects[0].properties["ref"].objects[1].metadata.uuid == uuid_A2

    C = client.collections.create(
        name="C",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            ReferenceProperty(name="ref", target_collection="B"),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C.data.insert({"Name": "find me", "ref": Reference.to(uuids=uuid_B)})

    objects = C.query.bm25(
        query="find",
        return_properties=[
            "name",
            FromReference(
                link_on="ref",
                return_properties=[
                    "name",
                    FromReference(
                        link_on="ref",
                        return_properties=["name"],
                        return_metadata=MetadataQuery(uuid=True),
                    ),
                ],
                return_metadata=MetadataQuery(uuid=True, last_update_time_unix=True),
            ),
        ],
    ).objects
    assert objects[0].properties["name"] == "find me"
    assert objects[0].properties["ref"].objects[0].properties["name"] == "B"
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[0].properties["name"]
        == "A1"
    )
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[1].properties["name"]
        == "A2"
    )


def test_mono_references_grpc_typed_dicts(client: weaviate.WeaviateClient):
    client.collections.delete("ATypedDicts")
    client.collections.delete("BTypedDicts")
    client.collections.delete("CTypedDicts")

    class AProps(TypedDict):
        name: str

    class BProps(TypedDict):
        name: str
        ref: CrossReference[AProps]

    class CProps(TypedDict):
        name: str
        ref: Annotated[CrossReference[BProps], MetadataQuery(uuid=True)]

    client.collections.create(
        name="ATypedDicts",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    A = client.collections.get("ATypedDicts", AProps)
    uuid_A1 = A.data.insert(AProps(name="A1"))
    uuid_A2 = A.data.insert(AProps(name="A2"))

    B = client.collections.create(
        name="BTypedDicts",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            ReferenceProperty(name="ref", target_collection="ATypedDicts"),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    B = client.collections.get("BTypedDicts", BProps)
    uuid_B = B.data.insert(
        properties=BProps(name="B", ref=Reference.to(uuids=uuid_A1, data_model=AProps))
    )
    B.data.reference_add(
        from_uuid=uuid_B,
        from_property="ref",
        ref=Reference.to(uuids=uuid_A2, data_model=AProps),
    )

    client.collections.create(
        name="CTypedDicts",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            Property(name="Age", data_type=DataType.INT),
            ReferenceProperty(name="ref", target_collection="BTypedDicts"),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C = client.collections.get("CTypedDicts", CProps)
    C.data.insert(
        properties=CProps(name="find me", ref=Reference.to(uuids=uuid_B, data_model=BProps))
    )

    objects = (
        client.collections.get("CTypedDicts")
        .query.bm25(
            query="find",
            return_properties=CProps,
        )
        .objects
    )
    assert (
        objects[0].properties["name"] == "find me"
    )  # happy path (in type and in return_properties)
    assert objects[0].metadata.uuid is None
    assert (
        objects[0].properties.get("not_specified") is None
    )  # type is str but instance is None (in type but not in return_properties)
    assert objects[0].properties["ref"].objects[0].properties["name"] == "B"
    assert objects[0].properties["ref"].objects[0].metadata.uuid == uuid_B
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[0].properties["name"]
        == "A1"
    )
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[0].metadata.uuid
        == uuid_A1
    )
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[1].properties["name"]
        == "A2"
    )
    assert (
        objects[0].properties["ref"].objects[0].properties["ref"].objects[1].metadata.uuid
        == uuid_A2
    )


def test_multi_references_grpc(client: weaviate.WeaviateClient):
    client.collections.delete("A")
    client.collections.delete("B")
    client.collections.delete("C")

    A = client.collections.create(
        name="A",
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
    )
    uuid_A = A.data.insert(properties={"Name": "A"})

    B = client.collections.create(
        name="B",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuid_B = B.data.insert({"Name": "B"})

    C = client.collections.create(
        name="C",
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            ReferencePropertyMultiTarget(name="ref", target_collections=["A", "B"]),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    C.data.insert(
        {
            "Name": "first",
            "ref": Reference.to_multi_target(uuids=uuid_A, target_collection="A"),
        }
    )
    C.data.insert(
        {
            "Name": "second",
            "ref": Reference.to_multi_target(uuids=uuid_B, target_collection="B"),
        }
    )

    objects = C.query.bm25(
        query="first",
        return_properties=[
            "name",
            FromReferenceMultiTarget(
                link_on="ref",
                target_collection="A",
                return_properties=["name"],
                return_metadata=MetadataQuery(uuid=True, last_update_time_unix=True),
            ),
        ],
    ).objects
    assert objects[0].properties["name"] == "first"
    assert len(objects[0].properties["ref"].objects) == 1
    assert objects[0].properties["ref"].objects[0].properties["name"] == "A"

    objects = C.query.bm25(
        query="second",
        return_properties=[
            "name",
            FromReferenceMultiTarget(
                link_on="ref",
                target_collection="B",
                return_properties=[
                    "name",
                ],
                return_metadata=MetadataQuery(uuid=True, last_update_time_unix=True),
            ),
        ],
    ).objects
    assert objects[0].properties["name"] == "second"
    assert len(objects[0].properties["ref"].objects) == 1
    assert objects[0].properties["ref"].objects[0].properties["name"] == "B"

    client.collections.delete("A")
    client.collections.delete("B")
    client.collections.delete("C")


def test_references_batch(client: weaviate.WeaviateClient):
    name_ref_to = "TestBatchRefTo"
    name_ref_from = "TestBatchRefFrom"

    client.collections.delete(name_ref_to)
    client.collections.delete(name_ref_from)

    ref_collection = client.collections.create(
        name=name_ref_to,
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[Property(name="num", data_type=DataType.INT)],
    )
    num_objects = 10

    uuids_to = ref_collection.data.insert_many(
        [DataObject(properties={"num": i}) for i in range(num_objects)]
    ).uuids.values()
    collection = client.collections.create(
        name=name_ref_from,
        properties=[
            ReferenceProperty(name="ref", target_collection=name_ref_to),
            Property(name="num", data_type=DataType.INT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )
    uuids_from = collection.data.insert_many(
        [DataObject(properties={"num": i}) for i in range(num_objects)]
    ).uuids.values()

    batch_return = collection.data.reference_add_many(
        [
            DataReference(
                from_property="ref", from_uuid=list(uuids_from)[i], to_uuid=list(uuids_to)[i]
            )
            for i in range(num_objects)
        ],
    )

    assert batch_return.has_errors is False

    objects = collection.query.fetch_objects(
        return_properties=[
            "num",
            FromReference(link_on="ref"),
        ],
    ).objects

    for obj in objects:
        assert obj.properties["num"] == obj.properties["ref"].objects[0].properties["num"]


def test_references_batch_with_errors(client: weaviate.WeaviateClient):
    name_ref_to = "TestBatchRefErrorTo"
    name_ref_from = "TestBatchRefErrorFrom"

    client.collections.delete(name_ref_to)
    client.collections.delete(name_ref_from)

    _ = client.collections.create(
        name=name_ref_to,
        vectorizer_config=Configure.Vectorizer.none(),
    )

    collection = client.collections.create(
        name=name_ref_from,
        properties=[
            ReferenceProperty(name="ref", target_collection=name_ref_to),
            Property(name="num", data_type=DataType.INT),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    batch_return = collection.data.reference_add_many(
        [DataReference(from_property="doesNotExist", from_uuid=uuid.uuid4(), to_uuid=uuid.uuid4())],
    )
    assert batch_return.has_errors is True
    assert 0 in batch_return.errors
    assert (
        batch_return.errors[0].message
        == "property doesNotExist does not exist for class TestBatchRefErrorFrom"
    )


def test_references_with_string_syntax(client: weaviate.WeaviateClient):
    name1 = "TestReferencesWithStringSyntaxA"
    name2 = "TestReferencesWithStringSyntaxB"
    client.collections.delete(name1)
    client.collections.delete(name2)

    client.collections.create(
        name=name1,
        vectorizer_config=Configure.Vectorizer.none(),
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            Property(name="Age", data_type=DataType.INT),
            Property(name="Weird__Name", data_type=DataType.INT),
        ],
    )

    uuid_A = client.collections.get(name1).data.insert(
        properties={"Name": "A", "Age": 1, "Weird__Name": 2}
    )

    client.collections.get(name1).query.fetch_object_by_id(uuid_A)

    client.collections.create(
        name=name2,
        properties=[
            Property(name="Name", data_type=DataType.TEXT),
            ReferenceProperty(name="ref", target_collection=name1),
        ],
        vectorizer_config=Configure.Vectorizer.none(),
    )

    client.collections.get(name2).data.insert({"Name": "B", "ref": Reference.to(uuids=uuid_A)})

    objects = (
        client.collections.get(name2)
        .query.bm25(
            query="B",
            return_properties=[
                "name",
                "__ref__properties__Name",
                "__ref__properties__Age",
                "__ref__properties__Weird__Name",
                "__ref__metadata__uuid",
                "__ref__metadata__last_update_time_unix",
            ],
        )
        .objects
    )

    assert objects[0].properties["name"] == "B"
    assert objects[0].properties["ref"].objects[0].properties["name"] == "A"
    assert objects[0].properties["ref"].objects[0].properties["age"] == 1
    assert objects[0].properties["ref"].objects[0].properties["weird__Name"] == 2
    assert objects[0].properties["ref"].objects[0].metadata.uuid == uuid_A
    assert objects[0].properties["ref"].objects[0].metadata.last_update_time_unix is not None
